
import os, sys
import time
import copy
import numpy as np
import healpy as hp
import itertools
import glob
import astropy, astroplan

from scipy.stats import norm
from scipy.optimize import minimize

import requests
import urllib.parse
import re

import astropy.coordinates
from astropy.time import Time, TimeDelta
import astropy.units as u
from astropy.coordinates import SkyCoord

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.path

import ligo.segments as segments
import ligo.skymap.distance as ligodist

import gwemopt.moc
import gwemopt.tiles
import gwemopt.segments
import gwemopt.scheduler

def readParamsFromFile(file):
    """@read gwemopt params file

    @param file
        gwemopt params file
    """

    params = {}
    if os.path.isfile(file):
        with open(file,'r') as f:
            for line in f:
                line_without_return = line.split("\n")
                line_split = line_without_return[0].split(" ")
                line_split = list(filter(None, line_split))
                if line_split:
                    try:
                        params[line_split[0]] = float(line_split[1])
                    except:
                        params[line_split[0]] = line_split[1]
    return params

def params_checker(params):
    """"Assigns defaults to params."""
    do_Parameters = ["do3D","doEvent","doSuperSched","doMovie_supersched","doSkymap","doSamples","doCoverage","doSchedule","doPlots","doDatabase","doMovie","doTiles","doIterativeTiling","doMinimalTiling","doOverlappingScheduling","doPerturbativeTiling","doOrderByObservability","doCatalog","doUseCatalog","doCatalogDatabase","doObservability","doSkybrightness","doEfficiency","doCalcTiles","doTransients","doSingleExposure","doAlternatingFilters","doMaxTiles","doReferences","doChipGaps","doUsePrimary","doSplit","doParallel","writeCatalog","doFootprint","doBalanceExposure","doBlocks","doUpdateScheduler","doTreasureMap","doRASlice"]
 
    for parameter in do_Parameters:
        if parameter not in params.keys():
            params[parameter] = False

    if "skymap" not in params.keys():
        params["skymap"] = '../output/skymaps/G268556.fits'

    if "gpstime" not in params.keys():
        params["gpstime"] = 1167559936.0

    if "outputDir" not in params.keys():
        params["outputDir"] = "../output"

    if "tilingDir" not in params.keys():
        params["tilingDir"] = "../tiling"

    if "catalogDir" not in params.keys():
        params["catalogDir"] = "../catalogs"

    if "event" not in params.keys():
        params["event"] = "G268556"

    if "coverageFiles" not in params.keys():
        params["coverageFiles"] = "../data/ATLAS_GW170104.dat"

    if "telescopes" not in params.keys():
        params["telescopes"] = 'ATLAS'

    if type(params["telescopes"]) == str:
        params["telescopes"] = params["telescopes"].split(",")

    if "lightcurveFiles" not in params.keys():
        params["lightcurveFiles"] = "../lightcurves/Me2017_H4M050V20.dat"

    if "tilesType" not in params.keys():
        params["tilesType"] = "moc"

    if "scheduleType" not in params.keys():
        params["scheduleType"] = "greedy"

    if "timeallocationType" not in params.keys():
        params["timeallocationType"] = "powerlaw"

    if "Ninj" not in params.keys():
        params["Ninj"] = 1000

    if "Ndet" not in params.keys():
        params["Ndet"] = 1
    
    if "Ntiles" not in params.keys():
        params["Ntiles"] = 10

    if "Ntiles_cr" not in params.keys():
        params["Ntiles_cr"] = 0.70

    if "Dscale" not in params.keys():
        params["Dscale"] = 1.0

    if "nside" not in params.keys():
        params["nside"] = 256

    if "Tobs" not in params.keys():
        params["Tobs"] = np.array([0.0,1.0])

    if "powerlaw_c1" not in params.keys():
        params["powerlaw_c1"] = 0.9

    if "powerlaw_n" not in params.keys():
        params["powerlaw_n"] = 1.0

    if "powerlaw_dist_exp" not in params.keys():
        params["powerlaw_dist_exp"] = 0

    if "galaxies_FoV_sep" not in params.keys():
        params["galaxies_FoV_sep"] = 1.0

    if "footprint_ra" not in params.keys():
        params["footprint_ra"] = 30.0

    if "footprint_dec" not in params.keys():
        params["footprint_dec"] = 60.0

    if "footprint_radius" not in params.keys():
        params["footprint_radius"] = 10.0

    if "transientsFile" not in params.keys():
        params["transientsFile"] = "../transients/ps1_objects.csv"

    if "dt" not in params.keys():
        params["dt"] = 14.0

    if "galaxy_catalog" not in params.keys():
        params["galaxy_catalog"] = "GLADE"
    
    if "filters" not in params.keys():
        params["filters"] = ['r','g','r']

    if "exposuretimes" not in params.keys():
        params["exposuretimes"] = np.array([30.0,30.0,30.0])

    if "max_nb_tiles" not in params.keys():
        params["max_nb_tiles"] = np.array([-1,-1,-1])

    if "mindiff" not in params.keys():
        params["mindiff"] = 0.0

    if "airmass" not in params.keys():
        params["airmass"] = 2.5

    if "iterativeOverlap" not in params.keys():
        params["iterativeOverlap"] = 0.0

    if "maximumOverlap" not in params.keys():
        params["maximumOverlap"] = 1.0

    if "catalog_n" not in params.keys():
        params["catalog_n"] = 1.0

    if "galaxy_grade" not in params.keys():
        params["galaxy_grade"] = "S"

    if "splitType" not in params.keys():
        params["splitType"] = "regional"

    if "Nregions" not in params.keys():
        params["Nregions"] = 768

    if "configDirectory" not in params.keys():
        params["configDirectory"] = "../config/"

    if "Ncores" not in params.keys():
        params["Ncores"] = 4

    if "Nblocks" not in params.keys():
        params["Nblocks"] = 4

    if "unbalanced_tiles" not in params.keys():
        params["unbalanced_tiles"] = None

    if "treasuremap_token" not in params.keys():
        params["treasuremap_token"] = ""

    if "treasuremap_status" not in params.keys():
        params["treasuremap_status"] = ["planned","completed"]

    if "graceid" not in params.keys():
        params["graceid"] = "S190426c"

    if "raslice" not in params.keys():
        params["raslice"] = [0.0,24.0]

    if "config" not in params.keys():
        params["config"] = {}
        configFiles = glob.glob("%s/*.config"%params["configDirectory"])
        for configFile in configFiles:
            telescope = configFile.split("/")[-1].replace(".config","")
            if not telescope in params["telescopes"]: continue
            params["config"][telescope] = gwemopt.utils.readParamsFromFile(configFile)
            params["config"][telescope]["telescope"] = telescope
            if params["doSingleExposure"]:
                exposuretime = np.array(opts.exposuretimes.split(","),dtype=np.float)[0]
            
                nmag = np.log(exposuretime/params["config"][telescope]["exposuretime"]) / np.log(2.5)
                params["config"][telescope]["magnitude"] = params["config"][telescope]["magnitude"] + nmag
                params["config"][telescope]["exposuretime"] = exposuretime
            if "tesselationFile" in params["config"][telescope]:
                if not os.path.isfile(params["config"][telescope]["tesselationFile"]):
                    if params["config"][telescope]["FOV_type"] == "circle":
                        gwemopt.tiles.tesselation_spiral(params["config"][telescope])
                    elif params["config"][telescope]["FOV_type"] == "square":
                        gwemopt.tiles.tesselation_packing(params["config"][telescope])
                if params["tilesType"] == "galaxy":
                    params["config"][telescope]["tesselation"] = np.empty((3,))
                else:
                    params["config"][telescope]["tesselation"] = np.loadtxt(params["config"][telescope]["tesselationFile"],usecols=(0,1,2),comments='%')

            if "referenceFile" in params["config"][telescope]:
                refs = table.unique(table.Table.read(
                    params["config"][telescope]["referenceFile"],
                    format='ascii', data_start=2, data_end=-1)['field', 'fid'])
                reference_images =\
                    {group[0]['field']: group['fid'].astype(int).tolist()
                    for group in refs.group_by('field').groups}
                reference_images_map = {1: 'g', 2: 'r', 3: 'i'}
                for key in reference_images:
                    reference_images[key] = [reference_images_map.get(n, n)
                                             for n in reference_images[key]]
                params["config"][telescope]["reference_images"] = reference_images
                                                                     
            location = astropy.coordinates.EarthLocation(params["config"][telescope]["longitude"],params["config"][telescope]["latitude"],params["config"][telescope]["elevation"])
            observer = astroplan.Observer(location=location)
            params["config"][telescope]["observer"] = observer

    return params

def read_skymap(params,is3D=False,map_struct=None):

    header = []
    if map_struct is None:
        map_struct = {}

        if params["doDatabase"]:
            models = params["models"]
            localizations_all = models.Localization.query.all()
            localizations = models.Localization.query.filter_by(dateobs=params["dateobs"],localization_name=params["localization_name"]).all()
            if localizations == None:
                raise ValueError("No localization with dateobs=%s"%params["dateobs"])
            else:
                prob_data = localizations[0].healpix
                prob_data = prob_data / np.sum(prob_data)
                map_struct["prob"] = prob_data
    
                distmu = localizations[0].distmu
                distsigma = localizations[0].distsigma
                distnorm = localizations[0].distnorm
    
                if distmu is None:            
                    map_struct["distmu"] = None
                    map_struct["distsigma"] = None
                    map_struct["distnorm"] = None
                else:
                    map_struct["distmu"] = np.array(distmu)
                    map_struct["distsigma"] = np.array(distsigma)
                    map_struct["distnorm"] = np.array(distnorm)
                    is3D = True
        else:
            filename = params["skymap"]
        
            if is3D:
                healpix_data, header = hp.read_map(filename, field=(0,1,2,3), verbose=False,h=True)
        
                distmu_data = healpix_data[1]
                distsigma_data = healpix_data[2]
                prob_data = healpix_data[0]
                norm_data = healpix_data[3]
        
                map_struct["distmu"] = distmu_data / params["DScale"]
                map_struct["distsigma"] = distsigma_data / params["DScale"]
                map_struct["prob"] = prob_data
                map_struct["distnorm"] = norm_data
    
            else:
                prob_data, header = hp.read_map(filename, field=0, verbose=False,h=True)
                prob_data = prob_data / np.sum(prob_data)
        
                map_struct["prob"] = prob_data

    natural_nside = hp.pixelfunc.get_nside(map_struct["prob"])
    nside = params["nside"]
    
    print("natural_nside =", natural_nside)
    print("nside =", nside)
    
    if not is3D:
        map_struct["prob"] = hp.ud_grade(map_struct["prob"],nside,power=-2)

    if is3D:
        if natural_nside != nside:
            map_struct["prob"], map_struct["distmu"],\
            map_struct["distsigma"], map_struct["distnorm"] = ligodist.ud_grade(map_struct["prob"],\
                                                                                map_struct["distmu"],\
                                                                                map_struct["distsigma"],\
                                                                                nside)
        
        nside_down = 32
        _, distmu_down,\
        distsigma_down, distnorm_down = ligodist.ud_grade(map_struct["prob"],\
                                                          map_struct["distmu"],\
                                                          map_struct["distsigma"],\
                                                          nside_down)

        r = np.linspace(0, 2000)
        map_struct["distmed"] = np.zeros(distmu_down.shape)
        for ipix in range(len(map_struct["distmed"])):
            dp_dr = r**2 * distnorm_down[ipix] * norm(distmu_down[ipix],distsigma_down[ipix]).pdf(r)
            dp_dr_norm = np.cumsum(dp_dr / np.sum(dp_dr))
            idx = np.argmin(np.abs(dp_dr_norm-0.5))
            map_struct["distmed"][ipix] = r[idx]
        map_struct["distmed"] = hp.ud_grade(map_struct["distmu"],nside,power=-2)

    npix = hp.nside2npix(nside)
    theta, phi = hp.pix2ang(nside, np.arange(npix))
    ra = np.rad2deg(phi)
    dec = np.rad2deg(0.5*np.pi - theta)

    map_struct["ra"] = ra
    map_struct["dec"] = dec

    if params["doRASlice"]:
        ra_low, ra_high = params["raslice"][0], params["raslice"][1]
        ipix = np.where((ra_high*360.0/24.0 < ra) | (ra_low*360.0/24.0 > ra))[0]
        map_struct["prob"][ipix] = 0.0
        map_struct["prob"] = map_struct["prob"] / np.sum(map_struct["prob"])

    sort_idx = np.argsort(map_struct["prob"])[::-1]
    csm = np.empty(len(map_struct["prob"]))
    csm[sort_idx] = np.cumsum(map_struct["prob"][sort_idx])

    map_struct["cumprob"] = csm
    map_struct["ipix_keep"] = np.where(csm <= params["iterativeOverlap"])[0]

    pixarea = hp.nside2pixarea(nside)
    pixarea_deg2 = hp.nside2pixarea(nside, degrees=True)

    map_struct["nside"] = nside
    map_struct["npix"] = npix
    map_struct["pixarea"] = pixarea
    map_struct["pixarea_deg2"] = pixarea_deg2
    
    for j in range(len(header)):
        if header[j][0] == "DATE":
            map_struct["trigtime"] = header[j][1]

    return map_struct   

def samples_from_skymap(map_struct, is3D = False, Nsamples = 100):

    prob_data_sorted = np.sort(map_struct["prob"])[::-1]
    prob_data_indexes = np.argsort(map_struct["prob"])[::-1]
    prob_data_cumsum = np.cumsum(prob_data_sorted)

    rand_values = np.random.rand(Nsamples,)

    ras = []
    decs = []
    dists = []

    if is3D:
        r = np.linspace(0, 2000)
        rand_values_dist = np.random.rand(Nsamples,)

    for ii in range(Nsamples):
        ipix = np.argmin(np.abs(prob_data_cumsum-rand_values[ii]))
        ra_inj = map_struct["ra"][prob_data_indexes][ipix]
        dec_inj = map_struct["dec"][prob_data_indexes][ipix]

        ras.append(ra_inj)
        decs.append(dec_inj)    

        if is3D:
            dp_dr = r**2 * map_struct["distnorm"][prob_data_indexes][ipix] * norm(map_struct["distmu"][prob_data_indexes][ipix], map_struct["distsigma"][prob_data_indexes][ipix]).pdf(r)
            dp_dr_norm = np.cumsum(dp_dr / np.sum(dp_dr))
            idx = np.argmin(np.abs(dp_dr_norm-rand_values_dist[ii]))
            dist_inj = r[idx]
            dists.append(dist_inj)
        else:
            dists.append(50.0)

    samples_struct = {}
    samples_struct["ra"] = np.array(ras)
    samples_struct["dec"] = np.array(decs)
    samples_struct["dist"] = np.array(dists)

    return samples_struct

def get_ellipse_coords(a=0.0, b=0.0, x=0.0, y=0.0, angle=0.0, npts=10):
    """ Draws an ellipse using (360*k + 1) discrete points; based on pseudo code
    given at http://en.wikipedia.org/wiki/Ellipse
    k = 1 means 361 points (degree by degree)
    a = major axis distance,
    b = minor axis distance,
    x = offset along the x-axis
    y = offset along the y-axis
    angle = clockwise rotation [in degrees] of the ellipse;
        * angle=0  : the ellipse is aligned with the positive x-axis
        * angle=30 : rotated 30 degrees clockwise from positive x-axis
    """
    pts = np.zeros((npts, 2))

    beta = -angle * np.pi/180.0
    sin_beta = np.sin(beta)
    cos_beta = np.cos(beta)
    alpha = np.linspace(0,2*np.pi,npts) 

    sin_alpha = np.sin(alpha)
    cos_alpha = np.cos(alpha)
    
    pts[:, 0] = x + (a * cos_alpha * cos_beta - b * sin_alpha * sin_beta)
    pts[:, 1] = y + (a * cos_alpha * sin_beta + b * sin_alpha * cos_beta)

    return pts

def getCirclePixels(ra_pointing, dec_pointing, radius, nside, alpha=0.4, color='k', edgecolor='k', rotation=None):

    theta = 0.5 * np.pi - np.deg2rad(dec_pointing)
    phi = np.deg2rad(ra_pointing)

    xyz = hp.ang2vec(theta, phi)
    ipix = hp.query_disc(nside, xyz, np.deg2rad(radius))

    radecs = get_ellipse_coords(a=radius/np.cos(np.deg2rad(dec_pointing)), b=radius, x=ra_pointing, y=dec_pointing, angle=0.0, npts=25)
    idx = np.where(radecs[:,1] > 90.0)[0]
    radecs[idx,1] = 180.0 - radecs[idx,1] 
    idx = np.where(radecs[:,1] < -90.0)[0]
    radecs[idx,1] = -180.0 - radecs[idx,1]    
    idx = np.where(radecs[:,0] > 360.0)[0]
    radecs[idx,0] = 720.0 - radecs[idx,0]
    idx = np.where(radecs[:,0] < 0.0)[0]
    radecs[idx,0] = 360.0 + radecs[idx,0]

    radecs = np.array(radecs)
    idx1 = np.where(radecs[:,0]>=180.0)[0]
    idx2 = np.where(radecs[:,0]<180.0)[0]
    idx3 = np.where(radecs[:,0]>300.0)[0]
    idx4 = np.where(radecs[:,0]<60.0)[0]
    if (len(idx1)>0 and len(idx2)>0) and not (len(idx3)>0 and len(idx4)>0):
        alpha = 0.0

    xyz = hp.ang2vec(radecs[:,0],radecs[:,1],lonlat=True)

    proj = hp.projector.MollweideProj(rot=rotation, coord=None)
    x,y = proj.vec2xy(xyz[:,0],xyz[:,1],xyz[:,2])
    xy = np.zeros(radecs.shape)
    xy[:,0] = x
    xy[:,1] = y
    #path = matplotlib.path.Path(xyz[:,1:3])
    path = matplotlib.path.Path(xy)
    patch = matplotlib.patches.PathPatch(path, alpha=alpha, color=color, fill=True, zorder=3, edgecolor=edgecolor)

    area = np.pi * radius**2

    return ipix, radecs, patch, area

def getRectanglePixels(ra_pointing, dec_pointing, raSide, decSide, nside, alpha = 0.4, color='k', edgecolor='k', rotation=None):

    area = raSide*decSide

    decCorners = (dec_pointing - decSide / 2.0, dec_pointing + decSide / 2.0)

    #security for the periodic limit conditions 
    radecs = []
    for d in decCorners:
        if d > 90.:
            d = 180. - d
        elif d < -90.:
            d = -180 - d

        raCorners = (ra_pointing - (raSide / 2.0) / np.cos(np.deg2rad(d)) , ra_pointing + (raSide / 2.0) / np.cos(np.deg2rad(d)))
        #security for the periodic limit conditions 
        for r in raCorners:
            if r > 360.:
                r = r - 360.
            elif r < 0.:
                r = 360. + r
            radecs.append([r,d])

    radecs = np.array(radecs)
    idx1 = np.where(radecs[:,0]>=180.0)[0]
    idx2 = np.where(radecs[:,0]<180.0)[0]
    idx3 = np.where(radecs[:,0]>300.0)[0]
    idx4 = np.where(radecs[:,0]<60.0)[0]
    if (len(idx1)>0 and len(idx2)>0) and not (len(idx3)>0 and len(idx4)>0):
        alpha = 0.0

    idx1 = np.where(np.abs(radecs[:,1])>=87.0)[0]
    if len(idx1) == 4:
        return [], [], [], []

    idx1 = np.where((radecs[:,1]>=87.0) | (radecs[:,1]<=-87.0))[0]
    if len(idx1)>0:
        radecs = np.delete(radecs, idx1[0], 0)

    xyz = []
    for r, d in radecs:
        xyz.append(hp.ang2vec(r, d, lonlat=True))

    npts, junk = radecs.shape
    if npts == 4:
        xyz = [xyz[0], xyz[1],xyz[3], xyz[2]]
        ipix = hp.query_polygon(nside, np.array(xyz))
    else:
        ipix = hp.query_polygon(nside, np.array(xyz))

    #idx1 = np.where((radecs[:,1]>=70.0) | (radecs[:,1]<=-70.0))[0]
    #idx2 = np.where((radecs[:,0]>300.0) | (radecs[:,0]<60.0))[0]
    #if (len(idx1) == 0) or (len(idx2) > 0):
    #    return [], [], [], []

    xyz = np.array(xyz)
    proj = hp.projector.MollweideProj(rot=rotation, coord=None)
    x,y = proj.vec2xy(xyz[:,0],xyz[:,1],xyz[:,2])
    xy = np.zeros(radecs.shape)
    xy[:,0] = x
    xy[:,1] = y
    path = matplotlib.path.Path(xy)
    patch = matplotlib.patches.PathPatch(path, alpha=alpha, color=color, fill=True, zorder=3, edgecolor=edgecolor)

    return ipix, radecs, patch, area

def getSquarePixels(ra_pointing, dec_pointing, tileSide, nside, alpha = 0.4, color='k', edgecolor='k', rotation=None):

    area = tileSide*tileSide

    decCorners = (dec_pointing - tileSide / 2.0, dec_pointing + tileSide / 2.0)

    #security for the periodic limit conditions 
    radecs = []
    for d in decCorners:
        if d > 90.:
            d = 180. - d
        elif d < -90.:
            d = -180 - d

        raCorners = (ra_pointing - (tileSide / 2.0) / np.cos(np.deg2rad(d)) , ra_pointing + (tileSide / 2.0) / np.cos(np.deg2rad(d)))

        #security for the periodic limit conditions 
        for r in raCorners:
            if r > 360.:
                r = r - 360.
            elif r < 0.:
                r = 360. + r
            radecs.append([r,d])

    radecs = np.array(radecs)
    idx1 = np.where(radecs[:,0]>=180.0)[0] 
    idx2 = np.where(radecs[:,0]<180.0)[0]
    idx3 = np.where(radecs[:,0]>300.0)[0]
    idx4 = np.where(radecs[:,0]<60.0)[0]
    if (len(idx1)>0 and len(idx2)>0) and not (len(idx3)>0 and len(idx4)>0):
        alpha = 0.0

    idx1 = np.where(np.abs(radecs[:,1])>=87.0)[0] 
    if len(idx1) == 4:
        return [], [], [], []

    idx1 = np.where((radecs[:,1]>=87.0) | (radecs[:,1]<=-87.0))[0]
    if len(idx1)>0:
        radecs = np.delete(radecs, idx1[0], 0)

    xyz = []
    for r, d in radecs:
        xyz.append(hp.ang2vec(r, d, lonlat=True))

    npts, junk = radecs.shape
    if npts == 4:
        xyz = [xyz[0], xyz[1],xyz[3], xyz[2]]
        ipix = hp.query_polygon(nside, np.array(xyz))
    else:
        ipix = hp.query_polygon(nside, np.array(xyz))

    #idx1 = np.where((radecs[:,1]>=70.0) | (radecs[:,1]<=-70.0))[0]
    #idx2 = np.where((radecs[:,0]>300.0) | (radecs[:,0]<60.0))[0]
    #if (len(idx1) == 0) or (len(idx2) > 0):
    #    return [], [], [], []

    xyz = np.array(xyz)
    proj = hp.projector.MollweideProj(rot=rotation, coord=None) 
    x,y = proj.vec2xy(xyz[:,0],xyz[:,1],xyz[:,2])
    xy = np.zeros(radecs.shape)
    xy[:,0] = x
    xy[:,1] = y
    path = matplotlib.path.Path(xy)
    patch = matplotlib.patches.PathPatch(path, alpha=alpha, color=color, fill=True, zorder=3, edgecolor=edgecolor)
    
    return ipix, radecs, patch, area

def integrationTime(T_obs, pValTiles, func=None, T_int=60.0):
    '''
    METHOD :: This method accepts the probability values of the ranked tiles, the 
              total observation time and the rank of the source tile. It returns 
              the array of time to be spent in each tile which is determined based
              on the localizaton probability of the tile. How the weight factor is 
              computed can also be supplied in functional form. Default is linear.
                      
    pValTiles :: The probability value of the ranked tiles. Obtained from ZTF_RT 
                             output
    T_obs     :: Total observation time available for the follow-up.
    func      :: functional form of the weight. Default is linear. 
                             For example, use x**2 to use a quadratic function.
    '''

    if func is None:
            f = lambda x: x
    else:
            f = lambda x: eval(func)
    fpValTiles = f(pValTiles)
    modified_prob = fpValTiles/np.sum(fpValTiles)
    modified_prob[np.isnan(modified_prob)] = 0.0
    t_tiles = modified_prob * T_obs ### Time spent in each tile if not constrained
    #t_tiles[t_tiles > 1200.0] = 1200.0 ### Upper limit of exposure time
    #t_tiles[t_tiles < 60] = 60.0 ### Lower limit of exposure time
    t_tiles = T_int*np.round(t_tiles/T_int)
    #Obs = np.cumsum(t_tiles) <= T_obs ### Tiles observable in T_obs seconds
    #time_per_tile = t_tiles[Obs] ### Actual time spent per tile

    return t_tiles

def observability(params, map_struct):

    airmass = params["airmass"]
    nside = params["nside"]
    npix = hp.nside2npix(nside)
    gpstime = params["gpstime"]
    event_time = Time(gpstime, format='gps', scale='utc')
    dts = np.arange(0,7,1.0/24.0)
    dts = np.arange(0,7,1.0/4.0)

    observatory_struct = {}

    for telescope in params["telescopes"]:
        config_struct = params["config"][telescope]

        observatory = astropy.coordinates.EarthLocation(
            lat=config_struct["latitude"]*u.deg, lon=config_struct["longitude"]*u.deg, height=config_struct["elevation"]*u.m)

        # Look up (celestial) spherical polar coordinates of HEALPix grid.
        theta, phi = hp.pix2ang(nside, np.arange(npix))
        # Convert to RA, Dec.
        radecs = astropy.coordinates.SkyCoord(
            ra=phi*u.rad, dec=(0.5*np.pi - theta)*u.rad)

        observatory_struct[telescope] = {}
        observatory_struct[telescope]["prob"] = copy.deepcopy(map_struct["prob"])
        observatory_struct[telescope]["observability"] = np.zeros((npix,))
        observatory_struct[telescope]["dts"] = {}

        for dt in dts:
            time = event_time+TimeDelta(dt*u.day)

            # Alt/az reference frame at observatory, now
            frame = astropy.coordinates.AltAz(obstime=time, location=observatory)
            # Transform grid to alt/az coordinates at observatory, now
            altaz = radecs.transform_to(frame)

            # Where is the sun, now?
            sun_altaz = astropy.coordinates.get_sun(time).transform_to(altaz)

            # How likely is it that the (true, unknown) location of the source
            # is within the area that is visible, now? Demand that sun is at
            # least 18 degrees below the horizon and that the airmass
            # (secant of zenith angle approximation) is at most 2.5.
            idx = np.where((altaz.alt >= 30*u.deg) &  (sun_altaz.alt <= -18*u.deg) & (altaz.secz <= airmass))[0]
            observatory_struct[telescope]["dts"][dt] = np.zeros((npix,))
            observatory_struct[telescope]["dts"][dt][idx] = 1
            observatory_struct[telescope]["observability"][idx] = 1
        observatory_struct[telescope]["prob"] = observatory_struct[telescope]["prob"]*observatory_struct[telescope]["observability"]

    return observatory_struct

def get_exposures(params, config_struct, segmentlist):
    '''
    Convert the availability times to a list segments with the length of telescope exposures.
    segmentlist: the segments that the telescope can do the follow-up.
    '''
    exposurelist = segments.segmentlist()
    if "overhead_per_exposure" in config_struct.keys(): overhead = config_struct["overhead_per_exposure"]
    else: overhead = 0.0

    # add the filter change time to the total overheads for integrated
    if not params["doAlternatingFilters"]: overhead = overhead + config_struct["filt_change_time"]

    exposure_time = np.max(params["exposuretimes"])

    for ii in range(len(segmentlist)):
        start_segment, end_segment = segmentlist[ii][0], segmentlist[ii][1]
        exposures = np.arange(start_segment, end_segment, (overhead+exposure_time)/86400.0)

        for jj in range(len(exposures)):
            exposurelist.append(segments.segment(exposures[jj],exposures[jj]+exposure_time/86400.0))

    return exposurelist

def perturb_tiles(params, config_struct, telescope, map_struct, tile_struct):

    map_struct_hold = copy.deepcopy(map_struct)
    ipix_keep = map_struct_hold["ipix_keep"]
    nside = params["nside"]

    if config_struct["FOV_type"] == "square":
        width = config_struct["FOV"]*0.5
    elif config_struct["FOV_type"] == "circle":
        width = config_struct["FOV"]*1.0

    moc_struct = {}
    keys = list(tile_struct.keys())
    for ii, key in enumerate(keys):
        if tile_struct[key]['prob'] == 0.0: continue

        if np.mod(ii,100) == 0:
            print("Optimizing tile %d/%d" % (ii, len(keys)))

        x0 = [tile_struct[key]["ra"], tile_struct[key]["dec"]]
        FOV = config_struct["FOV"]
        bounds = [[tile_struct[key]["ra"]-width, tile_struct[key]["ra"]+width],
                  [tile_struct[key]["dec"]-width, tile_struct[key]["dec"]+width]]

        ras = np.linspace(tile_struct[key]["ra"]-width, tile_struct[key]["ra"]+width, 5)
        decs = np.linspace(tile_struct[key]["dec"]-width, tile_struct[key]["dec"]+width, 5)
        RAs, DECs = np.meshgrid(ras, decs)
        ras, decs = RAs.flatten(), DECs.flatten()

        vals = []
        for ra, dec in zip(ras, decs):
            if np.abs(dec) > 90:
                vals.append(0)
                continue

            moc_struct_temp = gwemopt.moc.Fov2Moc(params, config_struct, telescope, ra, dec, nside)
            idx = np.where(map_struct_hold["prob"][moc_struct_temp["ipix"]] == -1)[0]
            idx = np.setdiff1d(idx,ipix_keep)
            if len(map_struct_hold["prob"][moc_struct_temp["ipix"]]) == 0:
                rat = 0.0
            else:
                rat = float(len(idx)) / float(len(map_struct_hold["prob"][moc_struct_temp["ipix"]]))
            if rat > params["maximumOverlap"]:
                val = 0.0
            else:
                ipix = moc_struct_temp["ipix"]
                if len(ipix) == 0:
                    val = 0.0
                else: 
                    vals_to_sum = map_struct_hold["prob"][ipix]
                    vals_to_sum[vals_to_sum < 0] = 0
                    val = np.sum(vals_to_sum) 
            vals.append(val)
        idx = np.argmax(vals)
        ra, dec = ras[idx], decs[idx]
        moc_struct[key] = gwemopt.moc.Fov2Moc(params, config_struct, telescope, ra, dec, nside)

        map_struct_hold['prob'][moc_struct[key]["ipix"]] = -1
        ipix_keep = np.setdiff1d(ipix_keep, moc_struct[key]["ipix"])

    tile_struct = gwemopt.tiles.powerlaw_tiles_struct(params, config_struct, telescope, map_struct, moc_struct)
    tile_struct = gwemopt.segments.get_segments_tiles(params, config_struct, tile_struct)
 
    return tile_struct

def slice_map_tiles(params, map_struct, coverage_struct):

    prob = copy.deepcopy(map_struct["prob"])
    prob[prob < 0] = 0.0

    sort_idx = np.argsort(prob)[::-1]
    csm = np.empty(len(prob))
    csm[sort_idx] = np.cumsum(prob[sort_idx])
    ipix_keep = np.where(csm <= params["iterativeOverlap"])[0]

    for ii in range(len(coverage_struct["ipix"])):
        data = coverage_struct["data"][ii,:]
        filt = coverage_struct["filters"][ii]
        ipix = coverage_struct["ipix"][ii]
        patch = coverage_struct["patch"][ii]
        FOV = coverage_struct["FOV"][ii]
        area = coverage_struct["area"][ii]

        observ_time, exposure_time, field_id, prob, airmass = data[2], data[4], data[5], data[6], data[7]

        ipix_slice = np.setdiff1d(ipix, ipix_keep)
        if len(ipix_slice) == 0: continue
        map_struct["prob"][ipix_slice] = -1

    return map_struct

def slice_number_tiles(params, telescope, tile_struct, coverage_struct):

    idx = params["telescopes"].index(telescope)
    max_nb_tile = params["max_nb_tiles"][idx]
    if max_nb_tile < 0:
        return tile_struct, False

    keys = tile_struct.keys()
    keys_scheduled = np.unique(coverage_struct["data"][:,5])
    
    if len(keys_scheduled) <= max_nb_tile:
        return tile_struct, False

    prob = np.zeros((len(keys),))
    for ii, key in enumerate(keys):
        if key in keys_scheduled:
            prob[ii] = prob[ii] + tile_struct[key]['prob']

    sort_idx = np.argsort(prob)[::-1]
    idx_keep = sort_idx[:int(max_nb_tile)]

    for ii, key in enumerate(keys):
        # in the golden tile set
        if ii in idx_keep: continue
        tile_struct[key]['prob'] = 0.0

    return tile_struct, True

def eject_tiles(params, telescope, tile_struct):
    
    for field_id in params["scheduled_fields"][telescope]:
        tile_struct[field_id]['prob'] = 0.0
    
    return tile_struct

def balance_tiles(params, tile_struct, coverage_struct):
    
    params["unbalanced_tiles"] = []
    
    filters, exposuretimes = params["filters"], params["exposuretimes"]

    keys_scheduled = coverage_struct["data"][:,5]
    filts = coverage_struct["filters"]
    filts_used = {key: [] for key in keys_scheduled} #dictionary of scheduled filters for each key

    for (key,filt) in zip(keys_scheduled,filts):
        filts_used[key].append(filt)

    doReschedule = False
    balanced_fields = 0
    for key in filts_used:
        if len(filts_used[key]) != len(filters):
            tile_struct[key]['prob'] = 0.0
            params["unbalanced_tiles"].append(key)
            doReschedule = True
        else:
            balanced_fields+=1
    return tile_struct, doReschedule, balanced_fields

def slice_galaxy_tiles(params, tile_struct, coverage_struct):

    coverage_ras = coverage_struct["data"][:,0]
    coverage_decs = coverage_struct["data"][:,1]

    if len(coverage_ras) == 0:
        return tile_struct

    keys = tile_struct.keys()
    ras, decs = [], []
    for key in keys:
        ras.append(tile_struct[key]["ra"])
        decs.append(tile_struct[key]["dec"])
    ras, decs = np.array(ras), np.array(decs)

    prob = np.zeros((len(keys),))
    for ii, key in enumerate(keys):
        prob[ii] = prob[ii] + tile_struct[key]['prob']

    sort_idx = np.argsort(prob)[::-1]
    csm = np.empty(len(prob))
    csm[sort_idx] = np.cumsum(prob[sort_idx])
    ipix_keep = np.where(csm <= params["iterativeOverlap"])[0]
 
    catalog1 = SkyCoord(ra=coverage_ras*u.degree,
                        dec=coverage_decs*u.degree, frame='icrs')

    for ii, key in enumerate(keys):
        # in the golden tile set
        if ii in ipix_keep: continue

        catalog2 = SkyCoord(ra=tile_struct[key]["ra"]*u.degree,
                            dec=tile_struct[key]["dec"]*u.degree,
                            frame='icrs')
        sep = catalog1.separation(catalog2)
        galaxies = tile_struct[key]["galaxies"]
        for jj, s in enumerate(sep):
            if s.deg > 1:
                continue
            galaxies2 = coverage_struct["galaxies"][jj]
            overlap = np.intersect1d(galaxies, galaxies2)
            if len(overlap) == 0:
                continue

            rat = np.array([float(len(overlap)) / float(len(galaxies)),
                            float(len(overlap)) / float(len(galaxies2))])

            if np.max(rat) > params["maximumOverlap"]:
                tile_struct[key]['prob'] = 0.0
                break

    return tile_struct

def optimize_max_tiles(params,tile_struct,coverage_struct,config_struct,telescope,map_struct_hold):
    """Returns value of max_tiles_nb optimized for number of scheduled fields with 'balanced' exposures in each filter."""
    prob={}
    for key in tile_struct.keys():
        prob[key] = tile_struct[key]['prob']
    
    keys_scheduled = coverage_struct["data"][:,5]
    filts_used = {key:[] for key in keys_scheduled}
    filts = coverage_struct["filters"]
    
    for (key,filt) in zip(keys_scheduled,filts):
        filts_used[key].append(filt)
    n_equal=0
    for key in filts_used:
        if len(filts_used[key]) == len(params["filters"]):
            n_equal+=1
    optimized_max=-1 #assigns baseline optimized maxtiles
    
    params["doMaxTiles"] = True
    countervals=[]
    
    coarse_bool = False
    repeating = False
    if (config_struct["FOV_type"] == "circle" and config_struct["FOV"] <= 2.0
        or config_struct["FOV_type"] == "square" and config_struct["FOV"] <= 4.0):
        max_trials = np.linspace(10,210,9)
        coarse_bool = True
    else:
        max_trials = np.linspace(10,200,20)

    for ii,max_trial in enumerate(max_trials):
        params["max_nb_tiles"] = np.array([max_trial],dtype=np.float)
        params_hold = copy.copy(params)
        tile_struct_hold = copy.copy(tile_struct)
        config_struct_hold = copy.copy(config_struct)
        
        coverage_struct_hold,tile_struct_hold = gwemopt.scheduler.schedule_alternating(params_hold, config_struct_hold, telescope, map_struct_hold, tile_struct_hold)

        keys_scheduled = coverage_struct_hold["data"][:,5]
        filts_used = {key:[] for key in keys_scheduled}
        filts = coverage_struct_hold["filters"]
        
        #count number of 'balanced' exposures
        counter=0
        for (key,filt) in zip(keys_scheduled,filts):
            filts_used[key].append(filt)
            if len(filts_used[key]) == len(params["filters"]):
                counter+=1
        countervals.append(counter)

        #update optimized value and check for breaking conditions
        if counter>=n_equal:
            n_equal,optimized_max = counter,max_trial
        if counter == n_equal and ii>2:
            repeating = countervals[ii] == countervals[ii-1] == countervals[ii-2]

        for key in tile_struct.keys():
            tile_struct[key]['prob'] = prob[key]
        if ii>0 and counter<countervals[ii-1] or repeating: break
        
    #optimize within narrower range for more precision
    if coarse_bool == True:
        max_trials = np.linspace(optimized_max-24,optimized_max+24,7)
    else:
        if optimized_max < 100:
            max_trials = np.linspace(optimized_max-9,optimized_max+9,10)
        else:
            max_trials = np.linspace(optimized_max-9,optimized_max+9,4)

    countervals=[]
    repeating = False
    for ii,max_trial in enumerate(max_trials):
        if optimized_max==-1: break #breaks if no max tiles restriction should be imposed
        params["max_nb_tiles"] = np.array([max_trial],dtype=np.float)
        params_hold = copy.copy(params)
        tile_struct_hold = copy.copy(tile_struct)
        config_struct_hold = copy.copy(config_struct)
        
        coverage_struct_hold,tile_struct_hold = gwemopt.scheduler.schedule_alternating(params_hold, config_struct_hold, telescope, map_struct_hold, tile_struct_hold)
        
        keys_scheduled = coverage_struct_hold["data"][:,5]
        filts_used = {key:[] for key in keys_scheduled}
        filts = coverage_struct_hold["filters"]
        
        counter=0
        for (key,filt) in zip(keys_scheduled,filts):
            filts_used[key].append(filt)
            if len(filts_used[key]) == len(params["filters"]):
                counter+=1
        countervals.append(counter)

        if counter>=n_equal:
            n_equal,optimized_max = counter,max_trial
        if counter == n_equal and ii > 1:
            repeating = countervals[ii] == countervals[ii-1] == countervals[ii-2]

        for key in tile_struct.keys():
            tile_struct[key]['prob'] = prob[key]
        if ii>0 and counter<countervals[ii-1] or repeating: break
    return optimized_max

def check_overlapping_tiles(params, tile_struct, coverage_struct):

    coverage_ras = coverage_struct["data"][:,0]
    coverage_decs = coverage_struct["data"][:,1]
#   coverage_mjds = coverage_struct["data"][:,2]
    coverage_ipixs = coverage_struct["ipix"]
    if len(coverage_ras) == 0:
        return tile_struct

    keys = list(tile_struct.keys())
    ras, decs = [], []
    for key in keys:
        ras.append(tile_struct[key]["ra"])
        decs.append(tile_struct[key]["dec"])
    ras, decs = np.array(ras), np.array(decs)

    catalog1 = SkyCoord(ra=coverage_ras*u.degree,
                        dec=coverage_decs*u.degree, frame='icrs')
    if params["tilesType"] == "galaxy":
        for ii, key in enumerate(keys):
            catalog2 = SkyCoord(ra=tile_struct[key]["ra"]*u.degree,
                                dec=tile_struct[key]["dec"]*u.degree,
                                frame='icrs')
            sep = catalog1.separation(catalog2)
            galaxies = tile_struct[key]["galaxies"]
            for jj, s in enumerate(sep):
                if s.deg > 1:
                    continue
                galaxies2 = coverage_struct["galaxies"][jj]
                overlap = np.setdiff1d(galaxies, galaxies2)
                if len(overlap) == 0:
                    if not 'epochs' in tile_struct[key]:
                        tile_struct[key]["epochs"] = np.empty((0,8))
                    tile_struct[key]["epochs"] = np.append(tile_struct[key]["epochs"],np.atleast_2d(coverage_struct["data"][jj,:]),axis=0)
    else:
        for ii, key in enumerate(keys):
            catalog2 = SkyCoord(ra=tile_struct[key]["ra"]*u.degree,
                                dec=tile_struct[key]["dec"]*u.degree,
                                frame='icrs')
            sep = catalog1.separation(catalog2)
            ipix = tile_struct[key]["ipix"]
            for jj, s in enumerate(sep):
                if s.deg > 25:
                    continue
                ipix2 = coverage_struct["ipix"][jj]
                overlap = np.intersect1d(ipix, ipix2)
                
                rat = np.array([float(len(overlap)) / float(len(ipix)),
                                float(len(overlap)) / float(len(ipix2))])
                                
                if len(overlap) == 0 or (params["doSuperSched"] and np.max(rat) < 0.50): continue
                if params["doSuperSched"]:
                    if 'epochs_telescope' not in tile_struct[key]:
                        tile_struct[key]["epochs_telescope"]=[]
                    tile_struct[key]["epochs_telescope"].append(coverage_struct["telescope"][jj])

                if not 'epochs' in tile_struct[key]:
                    tile_struct[key]["epochs"] = np.empty((0,8))
                    tile_struct[key]["epochs_overlap"] = []
                    tile_struct[key]["epochs_filters"]=[]
                    
                tile_struct[key]["epochs"] = np.append(tile_struct[key]["epochs"],np.atleast_2d(coverage_struct["data"][jj,:]),axis=0)
                tile_struct[key]["epochs_overlap"].append(len(overlap))
                tile_struct[key]["epochs_filters"].append(coverage_struct["filters"][jj])

    return tile_struct

def order_by_observability(params,tile_structs):
    observability = []
    for telescope in params["telescopes"]:
        config_struct = params["config"][telescope]
        tiles_struct = tile_structs[telescope]
        exposurelist = config_struct["exposurelist"]
        observability_prob = 0.0
        keys = tiles_struct.keys()
        for jj, key in enumerate(keys):
            tilesegmentlist = tiles_struct[key]["segmentlist"]
            if tiles_struct[key]["prob"] == 0: continue
            if tilesegmentlist.intersects_segment(exposurelist[0]):
                observability_prob = observability_prob + tiles_struct[key]["prob"]
        observability.append(observability_prob)
    idx = np.argsort(observability)[::-1]
    params["telescopes"] = [params["telescopes"][ii] for ii in idx]

def get_treasuremap_pointings(params):
    BASE = 'http://treasuremap.space/api/v0'
    TARGET = 'pointings'
    info = {
        "api_token": params["treasuremap_token"],
        "bands": params["filters"],
        "statuses": params["treasuremap_status"],
        "graceid": params["graceid"]
    }
    
    url = "{}/{}?{}".format(BASE, TARGET, urllib.parse.urlencode(info))

    try:
        r = requests.get(url = url)
    except requests.exceptions.RequestException as e:
        print(e)
        exit(1)

    observations = r.text.split("}")

    #dicts of instrument FOVs
    FOV_square = {44: 4.96, 47: 6.86}
    FOV_circle = {38: 1.1}
    
    #create coverage_struct
    coverage_struct = {}
    coverage_struct["data"] = np.empty((0, 8))
    coverage_struct["filters"] = []
    coverage_struct["ipix"] = []

    #read information into coverage struct
    for obs in observations:
        if "invalid api_token" in obs:
            print("Invalid Treasure Map API token.")
            exit(1)
        elif "POINT" not in obs: continue
        
        #get ra and dec
        pointing = re.search('\(([^)]+)', obs).group(1)
        pointing = pointing.split(" ")
        ra, dec = float(pointing[0]), float(pointing[1])
        
        #get filter
        filteridx = obs.find("band") + 10 #jump to starting index of filter
        filter = obs[filteridx:].split("\"")[0][:-1]

        #get instrument id
        instrumentidx = obs.find("instrumentid") + 16 #jump to starting index of instrument id
        instrument_id = int(obs[instrumentidx:].split(",")[0])

        #get ipix
        if instrument_id in FOV_square:
            ipix, radecs, patch, area = gwemopt.utils.getSquarePixels(ra, dec, FOV_square[instrument_id], params["nside"])
        elif instrument_id in FOV_circle:
            ipix, radecs, patch, area = gwemopt.utils.getCirclePixels(ra, dec, FOV_circle[instrument_id], params["nside"])
        else: continue
        
        coverage_struct["data"] = np.append(coverage_struct["data"], np.array([[ra, dec, -1, -1, -1, -1, -1, -1]]), axis=0)
        coverage_struct["filters"].append(filter)
        coverage_struct["ipix"].append(ipix)

    return coverage_struct





import os, sys
import copy

import numpy as np
import healpy as hp

from scipy.stats import norm

from astropy.time import Time

from shapely.geometry import MultiPoint

import gwemopt.utils
import gwemopt.rankedTilesGenerator
import gwemopt.samplers, gwemopt.segments
import gwemopt.quadrants
import gwemopt.moc

from gwemopt.segments import angular_distance

def get_rectangle(ras, decs, ra_size, dec_size):

    ras[ras>180.0] = ras[ras>180.0] - 360.0

    poly = MultiPoint([(x,y) for x,y in zip(ras,decs)]).envelope
    minx, miny, maxx, maxy = poly.bounds
    width = maxx - minx
    height = maxy - miny

    while (width < ra_size) or (height < dec_size):

        ra_mean, dec_mean = np.mean(ras), np.mean(decs)    
        dist = angular_distance(ra_mean, dec_mean,
                                ras, decs)
        idx = np.setdiff1d(np.arange(len(ras)),np.argmax(dist))
        ras, decs = ras[idx], decs[idx]

        if len(ras) == 1:
            return np.mod(ras[0], 360.0), decs[0]

        poly = MultiPoint([(x,y) for x,y in zip(ras,decs)]).envelope
        minx, miny, maxx, maxy = poly.bounds
        width = maxx - minx
        height = maxy - miny

    return np.mod((minx+maxx)/2.0, 360.0), (miny+maxy)/2.0    

def galaxy(params, map_struct, catalog_struct):
    nside = params["nside"]

    tile_structs = {}
    for telescope in params["telescopes"]:

        config_struct = params["config"][telescope]

        # Combine in a single pointing, galaxies that are distant by
        # less than FoV * params['galaxies_FoV_sep']
        # Take galaxy with highest proba at the center of new pointing
        FoV = params["config"][telescope]['FOV'] * params['galaxies_FoV_sep']
        if 'FOV_center' in params["config"][telescope]:
            FoV_center = params["config"][telescope]['FOV_center'] * params['galaxies_FoV_sep']
        else:
            FoV_center = params["config"][telescope]['FOV'] * params['galaxies_FoV_sep']

        new_ra = []
        new_dec = []
        new_Sloc = []
        new_Smass = []
        new_S = []
        galaxies = []
        idxRem = np.arange(len(catalog_struct["ra"])).astype(int)

        while len(idxRem) > 0:
            ii = idxRem[0]
            ra, dec, Sloc, S, Smass = catalog_struct["ra"][ii], catalog_struct["dec"][ii], catalog_struct["Sloc"][ii], catalog_struct["S"][ii], catalog_struct["Smass"][ii]    
            
            if config_struct["FOV_type"] == "square":
                decCorners = (dec - FoV, dec + FoV)
                # assume small enough to use average dec for corners
                raCorners = (ra - FoV/np.cos(np.deg2rad(dec)) , ra + FoV / np.cos(np.deg2rad(dec)))
                idx1 = np.where((catalog_struct["ra"][idxRem]>=raCorners[0]) & (catalog_struct["ra"][idxRem]<=raCorners[1]))[0]
                idx2 = np.where((catalog_struct["dec"][idxRem]>=decCorners[0]) & (catalog_struct["dec"][idxRem]<=decCorners[1]))[0]
                mask = np.intersect1d(idx1,idx2)

                if len(mask) > 1:
                    ra_center, dec_center = get_rectangle(catalog_struct["ra"][idxRem][mask], catalog_struct["dec"][idxRem][mask], FoV/np.cos(np.deg2rad(dec)), FoV)

                    decCorners = (dec_center - FoV/2.0, dec_center + FoV/2.0)
                    raCorners = (ra_center - FoV/(2.0*np.cos(np.deg2rad(dec))) , ra_center + FoV/(2.0*np.cos(np.deg2rad(dec))))
                    idx1 = np.where((catalog_struct["ra"][idxRem]>=raCorners[0]) & (catalog_struct["ra"][idxRem]<=raCorners[1]))[0]
                    idx2 = np.where((catalog_struct["dec"][idxRem]>=decCorners[0]) & (catalog_struct["dec"][idxRem]<=decCorners[1]))[0]
                    mask2 = np.intersect1d(idx1,idx2)

                    decCorners = (dec_center - FoV_center/2.0, dec_center + FoV_center/2.0)
                    raCorners = (ra_center - FoV_center/(2.0*np.cos(np.deg2rad(dec))) , ra_center + FoV_center/(2.0*np.cos(np.deg2rad(dec))))
                    idx1 = np.where((catalog_struct["ra"][idxRem]>=raCorners[0]) & (catalog_struct["ra"][idxRem]<=raCorners[1]))[0]
                    idx2 = np.where((catalog_struct["dec"][idxRem]>=decCorners[0]) & (catalog_struct["dec"][idxRem]<=decCorners[1]))[0]
                    mask3 = np.intersect1d(idx1,idx2)

                    # did the optimization help?
                    if (len(mask2) > 2) and (len(mask3) > 0):
                        mask = mask2
                else:
                    ra_center, dec_center = np.mean(catalog_struct["ra"][idxRem][mask]), np.mean(catalog_struct["dec"][idxRem][mask])

            elif config_struct["FOV_type"] == "circle":
                dist = angular_distance(ra, dec,
                                        catalog_struct["ra"][idxRem],
                                        catalog_struct["dec"][idxRem])
                mask = np.where((2 * FoV) >= dist)[0]
                if len(mask) > 1:
                    ra_center, dec_center = get_rectangle(catalog_struct["ra"][idxRem][mask], catalog_struct["dec"][idxRem][mask], (FoV/np.sqrt(2))/np.cos(np.deg2rad(dec)), FoV/np.sqrt(2))

                    dist = angular_distance(ra_center, dec_center,
                                            catalog_struct["ra"][idxRem],
                                            catalog_struct["dec"][idxRem])
                    mask2 = np.where(FoV >= dist)[0]
                    # did the optimization help?
                    if len(mask2) > 2:
                        mask = mask2
                else:
                    ra_center, dec_center = np.mean(catalog_struct["ra"][idxRem][mask]), np.mean(catalog_struct["dec"][idxRem][mask])

            new_ra.append(ra_center)
            new_dec.append(dec_center)
            new_Sloc.append(np.sum(catalog_struct["Sloc"][idxRem][mask]))
            new_S.append(np.sum(catalog_struct["S"][idxRem][mask]))
            new_Smass.append(np.sum(catalog_struct["Smass"][idxRem][mask]))
            galaxies.append(idxRem[mask])

            idxRem = np.setdiff1d(idxRem, idxRem[mask])

        # redefine catalog_struct
        catalog_struct_new = {}
        catalog_struct_new["ra"] = new_ra
        catalog_struct_new["dec"] = new_dec
        catalog_struct_new["Sloc"] = new_Sloc
        catalog_struct_new["S"] = new_S
        catalog_struct_new["Smass"] = new_Smass
        catalog_struct_new["galaxies"] = galaxies

        moc_struct = {}
        cnt = 0
        for ra, dec, Sloc, S, Smass in zip(catalog_struct_new["ra"], catalog_struct_new["dec"], catalog_struct_new["Sloc"], catalog_struct_new["S"], catalog_struct_new["Smass"]):
            moc_struct[cnt] = gwemopt.moc.Fov2Moc(params, config_struct, telescope, ra, dec, nside)
            cnt = cnt + 1
        
        tile_struct = powerlaw_tiles_struct(params, config_struct, telescope, map_struct, moc_struct)
        tile_struct = gwemopt.segments.get_segments_tiles(params, config_struct, tile_struct)

        cnt = 0
        for ra, dec, Sloc, S, Smass,  galaxies in zip(catalog_struct_new["ra"], catalog_struct_new["dec"], catalog_struct_new["Sloc"], catalog_struct_new["S"],catalog_struct_new["Smass"], catalog_struct_new["galaxies"]):
            if params["galaxy_grade"] == "Sloc":
                tile_struct[cnt]['prob'] = Sloc
            elif params["galaxy_grade"] == "S":
                tile_struct[cnt]['prob'] = S
            elif params["galaxy_grade"] == "Smass":
                tile_struct[cnt]['prob'] = Smass
                
            tile_struct[cnt]['galaxies'] = galaxies
            if config_struct["FOV_type"] == "square":
                tile_struct[cnt]['area'] = params["config"][telescope]['FOV']**2
            elif config_struct["FOV_type"] == "circle":
                tile_struct[cnt]['area'] = 4*np.pi*params["config"][telescope]['FOV']**2
            cnt = cnt + 1

        tile_structs[telescope] = tile_struct

    return tile_structs

def greedy(params, map_struct):

    tile_structs = {}
    for telescope in params["telescopes"]:
        config_struct = params["config"][telescope]

        tile_struct = gwemopt.samplers.greedy_tiles_struct(params, config_struct, telescope, map_struct)
        tile_struct = gwemopt.segments.get_segments_tiles(params, config_struct, tile_struct)

        tile_structs[telescope] = tile_struct

    return tile_structs

def hierarchical(params, map_struct):

    tile_structs = {}
    for telescope in params["telescopes"]:
        config_struct = params["config"][telescope]

        tile_struct = gwemopt.samplers.hierarchical_tiles_struct(params, config_struct, telescope, map_struct)
        tile_struct = gwemopt.segments.get_segments_tiles(params, config_struct, tile_struct)

        tile_structs[telescope] = tile_struct

    return tile_structs

def powerlaw_tiles_struct(params, config_struct, telescope, map_struct, tile_struct):

    keys = tile_struct.keys()
    ntiles = len(keys)
    if ntiles == 0:
        return tile_struct

    tot_obs_time = config_struct["tot_obs_time"]

    if "observability" in map_struct:
        prob = map_struct["observability"][telescope]["prob"]
    else:
        prob = map_struct["prob"]

    n, cl, dist_exp = params["powerlaw_n"], params["powerlaw_cl"], params["powerlaw_dist_exp"]
    
    if params["tilesType"] == "galaxy":
        tile_probs = compute_tiles_map(params, tile_struct, prob, func='center', ipix_keep=map_struct["ipix_keep"])
    else:
        tile_probs = compute_tiles_map(params, tile_struct, prob, func='np.sum(x)', ipix_keep=map_struct["ipix_keep"])

    tile_probs[tile_probs<np.max(tile_probs)*0.01] = 0.0
 
    prob_scaled = copy.deepcopy(prob)
    prob_sorted = np.sort(prob_scaled)[::-1]
    prob_indexes = np.argsort(prob_scaled)[::-1]
    prob_cumsum = np.cumsum(prob_sorted)
    index = np.argmin(np.abs(prob_cumsum - cl)) + 1
    #prob_scaled[prob_indexes[index:]] = 1e-10
    prob_scaled[prob_indexes[index:]] = 0.0
    prob_scaled = prob_scaled**n
    prob_scaled = prob_scaled / np.nansum(prob_scaled)
   
    if params["tilesType"] == "galaxy":
        ranked_tile_probs = compute_tiles_map(params, tile_struct, prob_scaled, func='center', ipix_keep=map_struct["ipix_keep"])
    else:
        ranked_tile_probs = compute_tiles_map(params, tile_struct, prob_scaled, func='np.sum(x)', ipix_keep=map_struct["ipix_keep"])

    ranked_tile_probs[np.isnan(ranked_tile_probs)] = 0.0
    ranked_tile_probs_thresh = np.max(ranked_tile_probs)*0.01
    ranked_tile_probs[ranked_tile_probs<=ranked_tile_probs_thresh] = 0.0
    ranked_tile_probs = ranked_tile_probs / np.nansum(ranked_tile_probs)

    if "distmed" in map_struct:
        distmed = map_struct["distmed"]
        distmed[distmed<=0] = np.nan
        distmed[~np.isfinite(distmed)] = np.nan
        #distmed[distmed<np.nanmedian(distmed)/4.0] = np.nanmedian(distmed)/4.0

        if params["tilesType"] == "galaxy":
            ranked_tile_distances = compute_tiles_map(params, tile_struct, distmed, func='center', ipix_keep=map_struct["ipix_keep"])
        else:
            ranked_tile_distances = compute_tiles_map(params, tile_struct, distmed, func='np.nanmedian(x)', ipix_keep=map_struct["ipix_keep"])        
        ranked_tile_distances_median = ranked_tile_distances / np.nanmedian(ranked_tile_distances)
        ranked_tile_distances_median = ranked_tile_distances_median**dist_exp
        ranked_tile_probs = ranked_tile_probs*ranked_tile_distances_median
        ranked_tile_probs = ranked_tile_probs / np.nansum(ranked_tile_probs)
        ranked_tile_probs[np.isnan(ranked_tile_probs)] = 0.0
        
    if params["doSingleExposure"]:
        keys = tile_struct.keys()
        ranked_tile_times = np.zeros((len(ranked_tile_probs),len(params["exposuretimes"])))
        for ii in range(len(params["exposuretimes"])):
            ranked_tile_times[ranked_tile_probs>0,ii] = params["exposuretimes"][ii]
        for key, prob, exposureTime, tileprob in zip(keys, ranked_tile_probs, ranked_tile_times, tile_probs):

            # Try to load the minimum duration of time from telescope config file
            # Otherwise set it to zero
            try:
                min_obs_duration = config_struct["min_observability_duration"] / 24
            except:
                min_obs_duration = 0.0

            # Check that a given tile is observable a minimum amount of time
            # If not set the proba associated to the tile to zero
            if 'segmentlist' and 'prob' in tile_struct[key] and tile_struct[key]['segmentlist'] and min_obs_duration > 0.0:
                observability_duration = 0.0 
                for counter in range(len(tile_struct[key]['segmentlist'])):
                    observability_duration += tile_struct[key]['segmentlist'][counter][1] - tile_struct[key]['segmentlist'][counter][0]
                if tile_struct[key]['prob'] > 0.0 and observability_duration < min_obs_duration: 
                   tileprob = 0.0

            tile_struct[key]["prob"] = tileprob
            if prob == 0.0:
                tile_struct[key]["exposureTime"] = 0.0
                tile_struct[key]["nexposures"] = 0
                tile_struct[key]["filt"] = []
            else:
                if params["doReferences"] and (telescope in ["ZTF", "DECam"]):
                    tile_struct[key]["exposureTime"] = []
                    tile_struct[key]["nexposures"] = []
                    tile_struct[key]["filt"] = []
                    if key in config_struct["reference_images"]:
                        for ii in range(len(params["filters"])):
                            if params["filters"][ii] in config_struct["reference_images"][key]:
                                tile_struct[key]["exposureTime"].append(exposureTime[ii])
                                tile_struct[key]["filt"].append(params["filters"][ii])
                        tile_struct[key]["nexposures"] = len(tile_struct[key]["exposureTime"])
                    else:
                        tile_struct[key]["exposureTime"] = 0.0
                        tile_struct[key]["nexposures"] = 0
                        tile_struct[key]["filt"] = []
                else:
                    tile_struct[key]["exposureTime"] = exposureTime
                    tile_struct[key]["nexposures"] = len(params["exposuretimes"])
                    tile_struct[key]["filt"] = params["filters"]
    else:
        ranked_tile_times = gwemopt.utils.integrationTime(tot_obs_time, ranked_tile_probs, func=None, T_int=config_struct["exposuretime"])

        keys = tile_struct.keys()

        for key, prob, exposureTime, tileprob in zip(keys, ranked_tile_probs, ranked_tile_times, tile_probs):
            # Try to load the minimum duration of time from telescope config file
            # Otherwise set it to zero
            try:
                min_obs_duration = config_struct["min_observability_duration"] / 24
            except:
                min_obs_duration = 0.0

            # Check that a given tile is observable a minimum amount of time
            # If not set the proba associated to the tile to zero
            if 'segmentlist' and 'prob' in tile_struct[key] and tile_struct[key]['segmentlist'] and min_obs_duration > 0.0:
                observability_duration = 0.0 
                for counter in range(len(tile_struct[key]['segmentlist'])):
                    observability_duration += tile_struct[key]['segmentlist'][counter][1] - tile_struct[key]['segmentlist'][counter][0]
                if tile_struct[key]['prob'] > 0.0 and observability_duration < min_obs_duration: 
                    prob = 0.0
 
            tile_struct[key]["prob"] = prob
            tile_struct[key]["exposureTime"] = exposureTime
            tile_struct[key]["nexposures"] = int(np.floor(exposureTime/config_struct["exposuretime"]))
            tile_struct[key]["filt"] = [config_struct["filt"]] * tile_struct[key]["nexposures"]
            
    return tile_struct

def moc(params, map_struct, moc_structs):

    tile_structs = {}
    for telescope in params["telescopes"]:
 
        config_struct = params["config"][telescope]
        moc_struct = moc_structs[telescope]

        tile_struct = powerlaw_tiles_struct(params, config_struct, telescope, map_struct, moc_struct)
        tile_struct = gwemopt.segments.get_segments_tiles(params, config_struct, tile_struct)
        tile_structs[telescope] = tile_struct

    return tile_structs

def pem_tiles_struct(params, config_struct, telescope, map_struct, tile_struct):

    tot_obs_time = config_struct["tot_obs_time"]

    if "observability" in map_struct:
        prob = map_struct["observability"][telescope]["prob"]
    else:
        prob = map_struct["prob"]

    ranked_tile_probs = compute_tiles_map(params, tile_struct, prob, func='np.sum(x)', ipix_keep=map_struct["ipix_keep"])
    ranked_tile_times = gwemopt.utils.integrationTime(tot_obs_time, ranked_tile_probs, func=None, T_int=config_struct["exposuretime"])

    lim_mag = config_struct["magnitude"]
    lim_time = config_struct["exposuretime"]

    tau = np.arange(lim_time,3600.0,lim_time)
    Loftau = None

    N_ref = 9.7847e9
    L_min = 4.9370e31
    L_max = 4.9370e33

    if "distmu" in map_struct:
        prob = map_struct["prob"]
        distnorm = map_struct["distnorm"]
        distmu = map_struct["distmu"]
        distsigma = map_struct["distsigma"]

        D_min = 1.0e7
        D_max = 1.0e10
        R = np.linspace(D_min/1e6, D_max/1e6)
        p_R = [np.sum(prob * rr**2 * distnorm * norm(distmu, distsigma).pdf(rr)) for rr in R]
        p_R = np.array(p_R)
        p_R = p_R / len(prob)

        R = R*1e6
        D_mu = None
        D_sig = None
    else:
        D_mu = 200.0e6
        D_sig = 60.0e6
        R = None
        p_R = None

    tau, prob = gwemopt.pem.Pem(lim_mag, lim_time, N_ref = N_ref, L_min = L_min, L_max = L_max, tau = tau, Loftau = Loftau, D_mu = D_mu, D_sig = D_sig, R = R, p_R = p_R)

    tprob, time_allocation = gwemopt.pem.Main(tot_obs_time, 0, 0, ranked_tile_probs, tau, prob, 'Eq')

    if params["doPlots"]:
        gwemopt.plotting.tauprob(params,tau,prob)

    keys = tile_struct.keys()
    for key, prob, exposureTime in zip(keys, ranked_tile_probs, time_allocation):
        tile_struct[key]["prob"] = prob
        tile_struct[key]["exposureTime"] = exposureTime
        tile_struct[key]["nexposures"] = int(np.floor(exposureTime/config_struct["exposuretime"]))
        tile_struct[key]["filt"] = [config_struct["filt"]] * tile_struct[key]["nexposures"]

    return tile_struct

def compute_tiles_map(params, tile_struct, skymap, func=None, ipix_keep=[]):

    if func is None:
        f = lambda x: np.sum(x)
        
    elif func == 'center':
        keys = tile_struct.keys()
        ntiles = len(keys)
        vals = np.nan*np.ones((ntiles,))
        nside = hp.npix2nside(len(skymap))
        for ii,key in enumerate(tile_struct.keys()):
            pix_center = hp.ang2pix(nside, tile_struct[key]['ra'], tile_struct[key]['dec'], lonlat=True)
            val = skymap[pix_center]
            vals[ii] = val        
        return vals
    
    else:
        f = lambda x: eval(func)

    prob = copy.deepcopy(skymap)

    keys = tile_struct.keys()
    ntiles = len(keys)
    vals = np.nan*np.ones((ntiles,))
    for ii,key in enumerate(tile_struct.keys()):
        idx = np.where(prob[tile_struct[key]["ipix"]] == -1)[0]
        idx = np.setdiff1d(idx,ipix_keep)
        if len(prob[tile_struct[key]["ipix"]]) == 0:
            rat = 0.0
        else:
            rat = float(len(idx)) / float(len(prob[tile_struct[key]["ipix"]]))
        if rat > params["maximumOverlap"]:
            vals[ii] = 0.0            
        else:
            ipix = tile_struct[key]["ipix"]
            if len(ipix) == 0:
                vals[ii] = 0.0
            else:
                vals_to_sum = prob[ipix]
                vals_to_sum[vals_to_sum < 0] = 0
                vals[ii] = f(vals_to_sum)

        ipix_keep = np.setdiff1d(ipix_keep, tile_struct[key]["ipix"])
        ipix_slice = np.setdiff1d(tile_struct[key]["ipix"],ipix_keep)
        if len(ipix_slice) > 0:
            prob[ipix_slice] = -1

    return vals

def tesselation_spiral(config_struct, scale=0.80):
    if config_struct["FOV_type"] == "square":
        FOV = config_struct["FOV"]*config_struct["FOV"]*scale
    elif config_struct["FOV_type"] == "circle":
        FOV = np.pi*config_struct["FOV"]*config_struct["FOV"]*scale

    area_of_sphere = 4*np.pi*(180/np.pi)**2
    n = int(np.ceil(area_of_sphere/FOV))
    print("Using %d points to tile the sphere..."%n)
 
    golden_angle = np.pi * (3 - np.sqrt(5))
    theta = golden_angle * np.arange(n)
    z = np.linspace(1 - 1.0 / n, 1.0 / n - 1, n)
    radius = np.sqrt(1 - z * z)
 
    points = np.zeros((n, 3))
    points[:,0] = radius * np.cos(theta)
    points[:,1] = radius * np.sin(theta)
    points[:,2] = z

    ra, dec = hp.pixelfunc.vec2ang(points, lonlat=True)
    fid = open(config_struct["tesselationFile"],'w')
    for ii in range(len(ra)):
        fid.write('%d %.5f %.5f\n'%(ii,ra[ii],dec[ii]))
    fid.close()   

def tesselation_packing(config_struct, scale=0.97):
    sphere_radius = 1.0
    if config_struct["FOV_type"] == "square":
        circle_radius = np.deg2rad(config_struct["FOV"]/2.0) * scale
    elif config_struct["FOV_type"] == "circle":
        circle_radius = np.deg2rad(config_struct["FOV"]) * scale
    vertical_count = int( (np.pi*sphere_radius)/(2*circle_radius) )

    phis = []
    thetas = []

    phi = -0.5*np.pi
    phi_step = np.pi/vertical_count
    while phi<0.5*np.pi:
        horizontal_count = int( (2*np.pi*np.cos(phi)*sphere_radius)/(2*circle_radius) )
        if horizontal_count==0: horizontal_count=1
        theta = 0
        theta_step = 2*np.pi/horizontal_count
        while theta<2*np.pi-1e-8:
            phis.append(phi)
            thetas.append(theta)
            theta += theta_step
        phi += phi_step
    dec = np.array(np.rad2deg(phis))
    ra = np.array(np.rad2deg(thetas))

    fid = open(config_struct["tesselationFile"],'w')
    for ii in range(len(ra)):
        fid.write('%d %.5f %.5f\n'%(ii,ra[ii],dec[ii]))
    fid.close()





import os
import requests
import urllib.parse
from astropy import time
import astropy.units as u
from astropy.table import Table
import numpy as np
import healpy as hp
import pandas as pd

import gwemopt.ztf_tiling

def ztf_queue():

    ZTF_URL = "http://127.0.0.1:9999"
    fields, quadrants = [], []    

    r = requests.get(
        urllib.parse.urljoin(ZTF_URL, 'current_queue'))
    data = r.json()
    queue = pd.read_json(data['queue'], orient='records')
    if len(queue) > 0:
        n_fields = len(queue['field_id'].unique())
        for field in queue['field_id'].unique():
            rcids = np.arange(64)
            for rcid in rcids:
                fields.append(int(field))
                quadrants.append(int(rcid))

    return np.array(fields), np.array(quadrants)


def ztf_depot(start_time=None, end_time=None):
    """ZTF depot reader.

    Ingests information about images from all program ids
    (including program_id = 1) based on the nightly summary.
    This supplements what is available from the TAP interface,
    where information about public images is not available.

    Parameters
    ----------
    start_time : astropy.Time
        Start time of request.
    end_time : astropy.Time)
        End time of request.

    """

    if start_time is None:
        start_time = time.Time.now() - time.TimeDelta(1.0*u.day)
    if end_time is None:
        end_time = time.Time.now()

    depotdir = 'https://ztfweb.ipac.caltech.edu/ztf/depot'

    fields, quadrants = [], []

    mjds = np.arange(np.floor(start_time.mjd), np.ceil(end_time.mjd))
    for mjd in mjds:
        this_time = time.Time(mjd, format='mjd')
        dstr = this_time.iso.split(" ")[0].replace("-", "")

        url = os.path.join(depotdir, '%s/goodsubs_%s.txt' % (dstr, dstr))
        deptable = get_ztf_depot_table(url)
        if len(deptable) == 0:
            continue
       
        obs_grouped_by_jd = deptable.group_by('jd').groups
        for jd, rows in zip(obs_grouped_by_jd.keys, obs_grouped_by_jd):
            obstime = time.Time(rows['jd'][0], format='jd').datetime
            for row in rows:
                if not row["programid"] == 1: continue
                fields.append(int(row['field']))
                quadrants.append(int(row['rcid']))

    return np.array(fields), np.array(quadrants)


def get_ztf_depot_table(url):
    with requests.get(url) as r:
        deptable = Table.read(r.text, format='ascii.fixed_width',
                              data_start=2, data_end=-1)
    return deptable


def get_skymap(params):
   
    nside = params["nside"]

    npix = hp.nside2npix(nside)
    theta, phi = hp.pix2ang(nside, np.arange(npix))
    ra = np.rad2deg(phi)
    dec = np.rad2deg(0.5*np.pi - theta)

    #fields, quadrants = ztf_queue()
    fields, quadrants = ztf_depot(start_time=params["start_time"],
                                  end_time=params["end_time"])

    tess = params["config"]["ZTF"]["tesselation"]
    ipixs = []
    for field in np.unique(fields):
        idx = np.where(tess[:,0] == field)[0]
        if len(idx) == 0:
            continue
        row = tess[idx[0],:]
        ra, dec = row[1], row[2]
        idx = np.where(field == fields)[0]
        quads = quadrants[idx]
        ipix = gwemopt.ztf_tiling.get_quadrant_ipix(nside, ra, dec,
                                                    subfield_ids=quads)
        if len(ipix) == 0: continue
        ipixs.append(list({y for x in ipix for y in x}))
    ipix = np.array(list({y for x in ipixs for y in x}))

    prob = np.zeros((npix,))
    prob[ipix] = 1.
    prob = prob / np.sum(prob)

    fitsfile = os.path.join(params["outputDir"],'skymap.fits')
    hp.fitsfunc.write_map(fitsfile,prob,overwrite=True)

    return fitsfile




import numpy as np

import matplotlib
#matplotlib.rc('text', usetex=True)
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 16})
matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
from matplotlib import path, patches
import matplotlib.pyplot as plt

from astropy.coordinates import SkyCoord, Angle
import astropy.units as u

from mocpy import MOC, WCS
from regions import CircleSkyRegion
from astroquery.vizier import Vizier
from astroquery.cds import cds

import gwemopt.utils

ra = 162.3277489
dec = 22.9827302

center = SkyCoord(ra, dec, unit='deg')
radius = Angle(10.0/3600.0, unit='deg')
cone = CircleSkyRegion(center, radius)

cats = cds.query_region(region=cone, fields=['ID', 'moc_sky_fraction', 'moc_access_url'])

catIDs = []
for cat in cats:
    catID = cat["ID"][4:]
    catIDs.append(catID)
catIDs = list(set(catIDs))

for catID in catIDs:
    result = Vizier.query_region(center, radius, catalog=catID)
    if len(result) == 0: continue
    result = result[0]
    print(catID,result)
print(cats["ID"])

moc_galex=cds.find_datasets(meta_data="ID=CDS/P/GALEXGR6/AIS/FUV", return_moc=True)

print(cats["moc_access_url"])
print(stop)

mocs = []
FOV = 6.86
for field_id, ra, dec in fields:
    print(field_id, ra, dec)
    ipix, radecs, patch, area = gwemopt.utils.getSquarePixels(
        ra, dec, FOV, nside)
    if len(radecs) == 4:
        idx = [0, 1, 3, 2]
        radecs = radecs[idx,:]
    idx1 = np.where(radecs[:,0]>=180.0)[0]
    idx2 = np.where(radecs[:,0]<180.0)[0]
    idx3 = np.where(radecs[:,0]>300.0)[0]
    idx4 = np.where(radecs[:,0]<60.0)[0]
    if (len(idx1)>0 and len(idx2)>0) and not (len(idx3)>0 and len(idx4)>0):
        continue
    try:
        #moc = MOC.from_healpix_cells(ipix, np.log2(nside)*np.ones(ipix.shape))
        moc = MOC.from_polygon(radecs[:,0]*u.degree, radecs[:,1]*u.degree,
                               max_depth=moc_order)
    except:
        continue

    reg = cds.query_region(region=moc)
    print(reg)
    print(stop)

    mocs.append(moc)

fig = plt.figure(figsize=(15, 10))
with WCS(fig,
          fov=330 * u.deg,
          center=SkyCoord(0, 0, unit='deg', frame='galactic'),
          coordsys='galactic',
          rotation=Angle(0, u.degree),
          projection="AIT") as wcs:
    ax = fig.add_subplot(1, 1, 1, projection=wcs)
    for moc in mocs:
        moc.fill(ax=ax, wcs=wcs, edgecolor='r', facecolor='r', linewidth=1.0, fill=True, alpha=0.5)
        moc.border(ax=ax, wcs=wcs, color="black", alpha=0.5)

plt.xlabel('ra')
plt.ylabel('dec')
plt.grid(color="black", linestyle="dotted")
plt.show()
plt.savefig('../output/moc/moc.pdf')
plt.close()




import h5py
import astropy.io.fits
from astropy.table import Table
from astropy import units as u

filename = '../catalog/CLU_20170106_galexwise_DaveUpdate.fits'
filename = '../catalog/CLU_20181213V2.fits'
filename = '../catalog/CLU_20190406_filled.hdf5'
filename = '../catalog/CLU_20190708_marshalFormat.hdf5'

t = Table.read(filename)
columns = ['name', 'ra', 'dec', 'distmpc', 'sfr_fuv', 'mstar',
           'magb', 'a', 'b2a', 'pa', 'btc']
t.keep_columns(columns)

t['ra'].unit = u.deg
t['dec'].unit = u.deg
t['distmpc'].unit = u.Mpc

t.write('../catalog/CLU.hdf5', path='/data', format='hdf5', serialize_meta=True, overwrite=True)

t2 = t[:10]

t2.write('../catalog/CLU_mini.hdf5', path='/data', format='hdf5', serialize_meta=True, overwrite=True)

print(t2)
print(stop)

hdul[1].data = hdul[1].data[:10]

names = hdul[1].data.name
ras, decs = hdul[1].data.ra, hdul[1].data.dec
sfr, mstar = hdul[1].data.sfr_fuv, hdul[1].data.mstar
distmpc, magb = hdul[1].data.distmpc, hdul[1].data.magb
a, b2a, pa = hdul[1].data.a, hdul[1].data.b2a, hdul[1].data.pa
btc = hdul[1].data.btc

t = Table([names, ras, decs, sfr, mstar, distmpc, magb,
           a, b2a, pa, btc],
          names=('names', 'ras', 'decs', 'sfr', 'mstar', 'distmpc', 'magb',
                 'a', 'b2a', 'pa', 'btc'))
t['ras'].unit = u.deg
t['decs'].unit = u.deg
t['distmpc'].unit = u.Mpc

#t.write('../catalog/CLU_astropy.hdf5', path='/data', format='hdf5', serialize_meta=True)

#t = Table.read('../catalog/CLU_astropy.hdf5')

with h5py.File('../catalog/CLU_mini.hdf5', 'w') as f:
    f.create_dataset('name', data=names)
    f.create_dataset('ra', data=ras)
    f.create_dataset('dec', data=decs)
    f.create_dataset('sfr_fuv', data=sfr)
    f.create_dataset('mstar', data=mstar)
    f.create_dataset('distmpc', data=distmpc)
    f.create_dataset('magb', data=magb)
    f.create_dataset('a', data=a)
    f.create_dataset('b2a', data=b2a)
    f.create_dataset('pa', data=pa)
    f.create_dataset('btc', data=btc)



import time
import copy

from mocpy import MOC
from astropy.table import Table
from joblib import Parallel, delayed
import healpy as hp
import numpy as np

import shapely.geometry

import gwemopt.utils
import gwemopt.tiles
import gwemopt.ztf_tiling
import gwemopt.decam_tiling

def create_moc(params, map_struct=None):

    nside = params["nside"]
    npix = hp.nside2npix(nside)

    if params["doMinimalTiling"]:
        prob = map_struct["prob"]

        n, cl, dist_exp = params["powerlaw_n"], params["powerlaw_cl"], params["powerlaw_dist_exp"]
        prob_scaled = copy.deepcopy(prob)
        prob_sorted = np.sort(prob_scaled)[::-1]
        prob_indexes = np.argsort(prob_scaled)[::-1]
        prob_cumsum = np.cumsum(prob_sorted)
        index = np.argmin(np.abs(prob_cumsum - cl)) + 1
        prob_indexes = prob_indexes[:index+1]

    if "doUsePrimary" in params:
        doUsePrimary = params["doUsePrimary"]
    else:
        doUsePrimary = False

    moc_structs = {}
    for telescope in params["telescopes"]:
        config_struct = params["config"][telescope]
        tesselation = config_struct["tesselation"]
        moc_struct = {}

        if params["doMinimalTiling"] and (config_struct["FOV"] < 1.0):
            idxs = hp.pixelfunc.ang2pix(map_struct["nside"], tesselation[:,1], tesselation[:,2], lonlat=True)
            isin = np.isin(idxs, prob_indexes)
  
            idxs = [i for i, x in enumerate(isin) if x] 
            print("Keeping %d/%d tiles" % (len(idxs), len(tesselation)))
            tesselation = tesselation[idxs,:]

        if params["doParallel"]:
            moclists = Parallel(n_jobs=params["Ncores"])(delayed(Fov2Moc)(params, config_struct, telescope, tess[1], tess[2], nside) for tess in tesselation)
            for ii, tess in enumerate(tesselation):
                index, ra, dec = tess[0], tess[1], tess[2]
                if (telescope == "ZTF") and doUsePrimary and (index > 880):
                    continue
                moc_struct[index] = moclists[ii]    
        else:
            for ii, tess in enumerate(tesselation):
                index, ra, dec = tess[0], tess[1], tess[2]
                if (telescope == "ZTF") and doUsePrimary and (index > 880):
                    continue
                index = index.astype(int)
                moc_struct[index] = Fov2Moc(params, config_struct, telescope, ra, dec, nside)

        if map_struct is not None:
             ipix_keep = map_struct["ipix_keep"]
        else:
            ipix_keep = []
        if params["doMinimalTiling"]:
            moc_struct_new = copy.copy(moc_struct)
            if params["tilesType"] == "galaxy":
                tile_probs = gwemopt.tiles.compute_tiles_map(params, moc_struct_new, prob, func='center', ipix_keep=ipix_keep)
            else:
                tile_probs = gwemopt.tiles.compute_tiles_map(params, moc_struct_new, prob, func='np.sum(x)', ipix_keep=ipix_keep)

            keys = moc_struct.keys()

            sort_idx = np.argsort(tile_probs)[::-1]
            csm = np.empty(len(tile_probs))
            csm[sort_idx] = np.cumsum(tile_probs[sort_idx])
            ipix_keep = np.where(csm <= cl)[0]

            probs = []
            moc_struct = {}
            cnt = 0
            for ii, key in enumerate(keys):
                if ii in ipix_keep:
                    moc_struct[key] = moc_struct_new[key]
                    cnt = cnt + 1

        moc_structs[telescope] = moc_struct

    return moc_structs

def Fov2Moc(params, config_struct, telescope, ra_pointing, dec_pointing, nside):
    """Return a MOC in fits file of a fov footprint.
       The MOC fov is displayed in real time in an Aladin plan.

       Input:
           ra--> right ascention of fov center [deg]
           dec --> declination of fov center [deg]
           fov_width --> fov width [deg]
           fov_height --> fov height [deg]
           nside --> healpix resolution; by default 
           """

    moc_struct = {}
   
    if "rotation" in params:
        rotation=params["rotation"]
    else:
        rotation=None
 
    if config_struct["FOV_type"] == "square": 
        ipix, radecs, patch, area = gwemopt.utils.getSquarePixels(ra_pointing, dec_pointing, config_struct["FOV"], nside, rotation=rotation)
    elif config_struct["FOV_type"] == "circle":
        ipix, radecs, patch, area = gwemopt.utils.getCirclePixels(ra_pointing, dec_pointing, config_struct["FOV"], nside, rotation=rotation)

    if params["doChipGaps"]:
        if telescope == "ZTF":
            ipixs = gwemopt.ztf_tiling.get_quadrant_ipix(nside, ra_pointing, dec_pointing)
            ipix = list({y for x in ipixs for y in x})
        elif telescope == "DECam":
            ipixs = gwemopt.decam_tiling.get_quadrant_ipix(nside, ra_pointing, dec_pointing)
            ipix = list({y for x in ipixs for y in x})
        #else:
        #    print("Requested chip gaps with non-ZTF detector, will use moc.")

    moc_struct["ra"] = ra_pointing
    moc_struct["dec"] = dec_pointing
    moc_struct["ipix"] = ipix
    moc_struct["corners"] = radecs
    moc_struct["patch"] = patch
    moc_struct["area"] = area

    if False:
    #if len(ipix) > 0:
        # from index to polar coordinates
        theta, phi = hp.pix2ang(nside, ipix)

        # converting these to right ascension and declination in degrees
        ra = np.rad2deg(phi)
        dec = np.rad2deg(0.5 * np.pi - theta)

        box_ipix = Table([ra, dec], names = ('RA[deg]', 'DEC[deg]'),
                 meta = {'ipix': 'ipix table'})

        moc_order = int(np.log(nside)/ np.log(2))
        moc = MOC.from_table( box_ipix, 'RA[deg]', 'DEC[deg]', moc_order )

        moc_struct["moc"] = moc
    else:
        moc_struct["moc"] = []

    return moc_struct




from astropy import time

start_time = time.Time('2019-04-25T08:18:05', format='isot', scale='utc')
end_time = time.Time('2019-04-26T12:06:21.930', format='isot', scale='utc')

start_time = time.Time(1240327333.348145, format='gps', scale='utc')
end_time = time.Time('2019-04-27T04:21:39.991', format='isot', scale='utc')

start_time = time.Time('2019-04-27T04:21:39.991', format='isot', scale='utc')
end_time = time.Time('2019-04-28T11:41:07.848', format='isot', scale='utc')

start_time = time.Time('2019-04-25T08:18:05', format='isot', scale='utc')
end_time = time.Time('2019-04-25T09:39:48', format='isot', scale='utc')

start_time = time.Time('2019-05-10T02:59:39.292', format='isot', scale='utc')
end_time = time.Time('2019-05-10T06:00:25', format='isot', scale='utc')

start_time = time.Time('2019-05-10T06:00:25', format='isot', scale='utc')
end_time = time.Time('2019-05-11T00:31:40', format='isot', scale='utc')

start_time = time.Time('2019-04-26T15:21:55', format='isot', scale='utc')
end_time = time.Time('2019-04-26T17:23:04.047', format='isot', scale='utc')

start_time = time.Time('2019-04-26T17:23:04.047', format='isot', scale='utc')
end_time = time.Time('2019-04-27T22:47:04.528', format='isot', scale='utc')

start_time = time.Time('2019-04-25T08:18:05', format='isot', scale='utc')
end_time = time.Time('2019-04-25T09:06:42.775', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-04-25T09:06:42.775', format='isot', scale='utc') 
end_time = time.Time('2019-04-25T12:51:34', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-04-25T12:51:34', format='isot', scale='utc')
end_time = time.Time('2019-04-25T14:51:57', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T21:32:46', format='isot', scale='utc')
end_time = time.Time('2019-08-15T04:15:06', format='isot', scale='utc')

start_time = time.Time('2019-08-14T21:10:39.013', format='isot', scale='utc')
end_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')

dt = end_time - start_time
print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T04:19:00', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-15T00:09:41', format='isot', scale='utc')
end_time = time.Time('2019-08-15T02:38:20', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T05:39:26', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T10:32:00', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-15T10:32:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T12:11:49.075', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time(58710.22978, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time(58710.22978, format='mjd', scale='utc')
end_time = time.Time(58710.44370, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time(58710.5, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time(58710.5, format='mjd', scale='utc')
end_time = time.Time(58711.5, format='mjd', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-14T23:11:00', format='isot', scale='utc')
end_time = time.Time('2019-08-15T08:07:00', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-08-15T12:40:37', format='isot', scale='utc')
end_time = time.Time('2019-08-15T15:13:38', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-09-01T23:31:01.838', format='isot', scale='utc')
end_time = time.Time('2019-09-01T23:37:46', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-09-01T23:37:46', format='isot', scale='utc')
end_time = time.Time('2019-09-04T05:45:10', format='isot', scale='utc')

dt = end_time - start_time
#print(float(dt.value)*24)

start_time = time.Time('2019-09-10T01:26:19.243', format='isot', scale='utc')
end_time = time.Time('2019-09-10T02:58:42.5', format='isot', scale='utc')

dt = end_time - start_time
print(float(dt.value)*24)

start_time = time.Time('2019-09-10T08:29:58.544', format='isot', scale='utc')
end_time = time.Time('2019-09-10T10:18:00', format='isot', scale='utc')

dt = end_time - start_time
print(float(dt.value)*24)



# WAW strategy implemented here: https://github.com/omsharansalafia/waw
# Original by Om Sharan Salafia
# Modified by Michael Coughlin

from __future__ import print_function

import numpy as np
import healpy as hp

from scipy.stats import norm

import gwemopt.lightcurve

def detectability_maps(params, t, map_struct, samples_struct=None, nside=256, verbose=False, limit_to_region=None):
    """
    Compute the detectability maps P(F(t)>Flim|ra,dec,...) for a given EM
    counterpart, at a given obs frequency, with a given flux limit, at
    given times after the merger.


    Parameters:
    -----------
    nu_obs: scalar
        Observing frequency in Hz.
    Flim: scalar
        Limiting flux density of the search in mJy.
    t: numpy 1D array
        The times [in days] at which the detectability maps must be computed.
    samples_struct: dictionary
        Dictionary containing the parameters of the posterior samples. The
        dictionary is passed to the counterpart function that computes/retrieves
        the lightcurves. The dictionary *must* contain at least the keys 'ra'
        and 'dec' that contain the sky positions of the posterior samples.
    counterpart: function
        The function that is used to compute/retrieve the counterpart lightcurves.
        It must comply with the input/output specifications defined in
        emcounterparts.IO_specifications.
    nside: int
        The nside parameter that defines the healpix resolution of the output
        detectability maps.
    verbose: boolean
        If True, print the advancement status of the lightcurve computation/retrieval.
    limit_to_region: numpy 1D array
        A healpix map of booleans. The detectability maps will be computed only
        at sky positions that correspond to True pixels in this map. If None,
        the computation will be all-sky.

    Returns:
    --------
    detmaps: ndarray
        The shape of the output ndarray is (nt,npix), where nt is the length
        of the input t array, and npix = 12*nside**2. In practice, detmaps[i]
        is the detectability map at time t[i]. Conversely, detmaps[:,p] is
        the time-dependent detectability at sky position p (p is the healpix
        index! The actual coordinates are given by hp.pix2ang(nside,p)).
    """

    lightcurve_structs = gwemopt.lightcurve.read_files(params["lightcurveFiles"])
    for key in lightcurve_structs.iterkeys():
        lightcurve_struct = lightcurve_structs[key]   

    # compute the lightcurve of the counterpart for each posterior sample
    #F = gwemopt.lightcurve.compute_apparent_magnitude_samples(params, lightcurve_struct, samples_struct, t)
    F = gwemopt.lightcurve.compute_apparent_magnitude(params, lightcurve_struct, t)
    Flim = params["config"][params["telescopes"][0]]["magnitude"] 

    # compute the detectability maps
    detmaps = np.empty([len(t), hp.nside2npix(nside)])

    if verbose:
        print("")
    for i in range(len(t)):
        if verbose:
            print(" #### computing detmap {0:d} of {1:d} ({2:.1f} percent completed) ####".format(i + 1, len(t),i / len(t) * 100.),end="\r")

        #detmaps[i] = sky_pos_cond_prob_gt(F[:, i], Flim, samples_struct['ra'], samples_struct['dec'], nside,limit_to_region)
        #detmaps[i] = sky_pos_cond_prob_gt(10**(F[:, i]/-2.5), 10**(Flim/-2.5), samples_struct['ra'], samples_struct['dec'], nside,limit_to_region)
        detmaps[i] = sky_pos_cond_prob(F[i], Flim, map_struct, nside, limit_to_region)

    if verbose:
        print(" #### {} detmaps computed (100 percent completed) ####".format(len(t)))

    return detmaps

def construct_followup_strategy_tiles(skymap, detmaps, t_detmaps, tile_struct, T_int, T_available, min_detectability=0.01):

    # make sure that the detectability is above the minimum at some point
    if np.all(detmaps[np.isfinite(detmaps)] < min_detectability):
        print("No point in the detmaps is above the minumum required detectability.")
        return None

    T = T_int
 
    ntiles = len(tile_struct.keys())
    ridx = np.arange(ntiles) # this keeps track of the original healpix indices corresponding to the region
    dm = np.empty([len(detmaps), ntiles])
    for i in range(len(detmaps)):
        dm[i] = gwemopt.tiles.compute_tiles_map(tile_struct, detmaps[i], func='np.mean(x)') 

    # bring the skymap to the same resolution, and take only the region
    sm = gwemopt.tiles.compute_tiles_map(tile_struct, skymap, func='np.sum(x)')
    # also, find the descending probability sorted indices of the skymap
    descending_prob_idx = np.argsort(sm)[::-1]

    # how much total observing time is available? Just sum the differences
    # between ending and starting times of the available time windows,
    # then convert to seconds
    n_windows = len(T_available) // 2
    tot_obs_time = np.sum(np.diff(T_available)[::2]) * 86400.
    # compute the number of time slots, and their starting times in days.
    # Note that the number of slots per window must be an integer, thus
    # a the end of each window there might be some remainder time unused
    slots_in_window = (np.diff(T_available)[::2] * 86400./T).astype(int)
    n_slots = np.sum(slots_in_window)
    t0_slot = np.empty(n_slots)
    k = 0
    for i in range(n_windows):
        for j in range(slots_in_window[i]):
            t0_slot[k] = T_available[2 * i] + j * T/86400.
            k = k + 1
    # mark all slots as available
    available_slots = np.ones(n_slots,dtype=bool)

    # the strategy will be a healpix map of observation times
    strategy = np.ones(len(sm))*np.nan

    # assign the available time slots to the skymap pixels in order
    # of descending sky position probability. Each pixel is assigned
    # the available time slot where the detectability is highest
    for p in descending_prob_idx:
        try: # find the time of best detectability
            detp = np.interp(t0_slot[available_slots],t_detmaps,dm[:,p])
            i_best = np.argmax(detp)
            i0_best = np.arange(len(available_slots))[available_slots][i_best] #original index of the best time slot
        except:
            continue

        # do not assign observation if the best detectability is below the requested limit
        if detp[i_best]<min_detectability:
            continue
        else:
            strategy[ridx[p]]=t0_slot[i0_best] # assign the observation time
            available_slots[i0_best]=False # mark the assigned slot as not available anymore

    strategy[np.isnan(strategy)] = 0.0
    return strategy
    #return np.ma.masked_invalid(strategy)

def construct_followup_strategy(skymap, detmaps, t_detmaps, Afov, T_int, T_available, min_detectability=0.01, limit_to_region=None):
    """
    Construct the EM follow-up as in Salafia+17.

    Parameters:
    -----------
    skymap: 1D numpy array
        The healpix map of GW posterior sky position probability per unit area.
    detmaps: ndarray
        The detectability maps as computed by ::function::detectability_maps.
    t_detmaps: 1D array
        The times (in days) corresponding to the detectability maps provided.
    Afov: scalar
        The field of view of the observing instrument in square degrees.
    T_int: scalar
        The exposure/integration time (in seconds) that corresponds to the
        search limiting flux.
    T_available: tuple
        The starting and ending times (in days after the merger) of the
        available time windows. Must be an even number! (each window has
        a starting time and an ending time).
    min_detectability: scalar
        The minimum value of the detectability for which an observation
        can be scheduled.
    ...

    Return:
    -------
    strategy: 1D masked numpy array
        A healpix map containing the observation times corresponding
        to the follow-up strategy

    """

    # make sure that the detectability is above the minimum at some point
    if np.all(detmaps[np.isfinite(detmaps)] < min_detectability):
        print("No point in the detmaps is above the minumum required detectability.")
        return None

    # determine the nside that makes the pixel area closest to the fov
    # area, but not larger (FIXME: a MOC based representation of the fov
    # would do much better!)
    nsides = 2. ** np.arange(2, 10)
    areas = hp.nside2pixarea(nsides, degrees=True)
    ratio = Afov / areas
    nside = np.min(nsides[ratio > 1.])

    # rescale the integration time to the effective one corresponding to
    # that pixel area
    Apixel = hp.nside2pixarea(nside, degrees=True)
    T = T_int * Apixel / Afov  # this is the time needed to cover one healpix pixel

    # bring the limit_to_region to the correct resolution
    if limit_to_region is None:
        region = np.ones(hp.nside2npix(nside),dtype=bool)
    else:
        #region = hp.ud_grade(limit_to_region, nside)
        limit_to_region_ones = np.zeros(len(limit_to_region))
        limit_to_region_ones[np.where(limit_to_region==True)[0]] = 1.0
        region = hp.ud_grade(limit_to_region_ones,nside)
        region = np.ceil(region)
        region = region.astype(bool)

    # limit the detmaps to the region (to save memory and computation time)
    # after degrading/upgrading to the correct resolution
    ridx = np.arange(hp.nside2npix(nside))[region] # this keeps track of the original healpix indices corresponding to the region

    dm = np.empty([len(detmaps), len(region[region])])

    for i in range(len(detmaps)):
        dm[i] = hp.ud_grade(detmaps[i], nside)[region]

    # bring the skymap to the same resolution, and take only the region
    sm = hp.ud_grade(skymap,nside)[region]
    # also, find the descending probability sorted indices of the skymap
    descending_prob_idx = np.argsort(sm)[::-1]

    # how much total observing time is available? Just sum the differences
    # between ending and starting times of the available time windows,
    # then convert to seconds
    n_windows = len(T_available) // 2
    tot_obs_time = np.sum(np.diff(T_available)[::2]) * 86400.
    # compute the number of time slots, and their starting times in days.
    # Note that the number of slots per window must be an integer, thus
    # a the end of each window there might be some remainder time unused
    slots_in_window = (np.diff(T_available)[::2] * 86400./T).astype(int)
    n_slots = np.sum(slots_in_window)
    t0_slot = np.empty(n_slots)
    k = 0
    for i in range(n_windows):
        for j in range(slots_in_window[i]):
            t0_slot[k] = T_available[2 * i] + j * T/86400.
            k = k + 1
    # mark all slots as available
    available_slots = np.ones(n_slots,dtype=bool)

    # the strategy will be a healpix map of observation times
    strategy = np.ones(hp.nside2npix(nside))*np.nan

    # assign the available time slots to the skymap pixels in order
    # of descending sky position probability. Each pixel is assigned
    # the available time slot where the detectability is highest
    for p in descending_prob_idx:
        try: # find the time of best detectability
            detp = np.interp(t0_slot[available_slots],t_detmaps,dm[:,p])
            i_best = np.argmax(detp)
            i0_best = np.arange(len(available_slots))[available_slots][i_best] #original index of the best time slot
        except:
            continue

        # do not assign observation if the best detectability is below the requested limit
        if detp[i_best]<min_detectability:
            continue
        else:
            strategy[ridx[p]]=t0_slot[i0_best] # assign the observation time
            available_slots[i0_best]=False # mark the assigned slot as not available anymore
 
    strategy[np.isnan(strategy)] = 0.0
    return strategy
    #return np.ma.masked_invalid(strategy)

def sky_pos_cond_prob(x,x0,map_struct,nside=256,limit_to_region=None):

        npix = hp.nside2npix(nside)
        # precompute mask that selects x_i>x0
        #gt_x0 = x>x0

        # create an empty map
        m = np.ones(npix)*np.nan

        # if no limit_to_region given, create a mask with True everywhere,
        # otherwise adapt the given region resolution to ours
        if limit_to_region is None:
                region = np.ones(npix,dtype=bool)
        else:
                limit_to_region_ones = np.zeros(len(limit_to_region))
                limit_to_region_ones[np.where(limit_to_region==True)[0]] = 1.0
                region = hp.ud_grade(limit_to_region_ones,nside)
                region = np.ceil(region)
                region = region.astype(bool)

        distmu = hp.ud_grade(map_struct["distmu"],nside)
        distnorm = hp.ud_grade(map_struct["distnorm"],nside) 
        distsigma = hp.ud_grade(map_struct["distsigma"],nside)

        # to compute the angular distances of all posterior samples to all pixels of the
        # skymap is quite memory intensive, so we better do it one skymap pixel at a time
        r = np.linspace(1, 2000, 2000)

        for p in np.arange(npix)[region]:
                dp_dr = r**2 * distnorm[p] * norm(distmu[p],distsigma[p]).pdf(r)
                dp_dr_norm = np.cumsum(dp_dr / np.sum(dp_dr))

                app_m = x + 5*(np.log10(r*1e6) - 1) 
                idx = np.argmin(np.abs(app_m-x0))
               
                m[p] = dp_dr_norm[idx]
        
        #return the map, masking invalid values
        m[np.isnan(m)] = 0.0
        return m

# sky_pos_cond_prob_gt, i.e. sky-position-conditional probability that x is greater than x0. 
# In a more compact form: P(x>x0|ra,dec)
def sky_pos_cond_prob_gt(x,x0,ra,dec,nside=256,limit_to_region=None):
	"""
	
	Return a numpy array which represents the healpix projection of the 
	sky-position-conditional probabilities that x>x0, i.e. P(x>x0|sky pos,...),
	constructed from a set of posterior samples as described in Salafia+17
	
	
	Parameters
	----------
	x: numpy 1D array
	    Value of x for each posterior sample.
	x0: float
	    Comparison value.
	ra, dec: numpy arrays of the same length as x
	    Sky positions of the posterior samples.
	nside: int, default=32
	    Resolution of the output healpix map.
	limit_to_region: 1D numpy array of booleans
	    A healpix map of booleans representing the region where the probability has to be computed.
	    Typically, one wants to limit the computation to the 90% confidence area not to waste
	    computational resources. If its nside is different from that given, it will be converted.
	    Default: None (i.e. all-sky).
	
	Returns
	-------
	m: a masked 1D numpy array of length 12*nside**2, which represents the output healpix map.
	"""
	
	# assert that x, ra and dec lengths match
	assert len(x) == len(ra)," The x and ra arrays must have the same lengths"
	assert len(x) == len(dec)," The x and dec arrays must have the same lengths"
	
	# compute the cartesian components of the vectors pointing at the positions
	# of the posterior samples
	npsamp = len(x)
	vx = np.cos(dec)*np.cos(ra)
	vy = np.cos(dec)*np.sin(ra)
	vz = np.sin(dec)
	
	# compute the cartesian components of the vectors pointing at the positions
	# of the skymap pixels
	npix = hp.nside2npix(nside)
	px,py,pz = hp.pix2vec(nside,np.arange(npix))
	
	# precompute mask that selects x_i>x0
	gt_x0 = x>x0

	# create an empty map
	m = np.ones(npix)*np.nan
	
	# if no limit_to_region given, create a mask with True everywhere,
	# otherwise adapt the given region resolution to ours
	if limit_to_region is None:
		region = np.ones(npix,dtype=bool)
	else:
		limit_to_region_ones = np.zeros(len(limit_to_region))
		limit_to_region_ones[np.where(limit_to_region==True)[0]] = 1.0
		
		region = hp.ud_grade(limit_to_region_ones,nside)
		region = np.ceil(region)
		region = region.astype(bool)

	# to compute the angular distances of all posterior samples to all pixels of the
	# skymap is quite memory intensive, so we better do it one skymap pixel at a time
	for p in np.arange(npix)[region]:
		# compute angular distances
		d = np.arccos(px[p]*vx + py[p]*vy + pz[p]*vz)
		
		#compute IDW weigths and normalize them
		bandwidth = np.std(d)*npsamp**(-1./5.) #Silverman's rule (FIXME: use a better estimate?)
		w = np.exp(-0.5*(d/bandwidth)**2)
	
		if np.sum(w)>0.: #do not normalize if the sum is 0
			w = w/np.sum(w)
	
		#compute the probability as prescribed in Salafia+17
		m[p] = np.sum(w[gt_x0])
	
	#return the map, masking invalid values
	m[np.isnan(m)] = 0.0
	return m
	#return np.ma.masked_invalid(m)



# Copyright (C) 2018 David Kaplan
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import print_function
import numpy as np
import healpy as hp
from astropy.table import Table
from astropy import constants as c, units as u
from astropy.coordinates import SkyCoord, ICRS
from astropy.wcs import WCS
import astropy
import healpy
# import cPickle as pickle
import pickle
import os


class ZTFtile:

    # make sure it can find the table of quadrant centers
    #classlocation=os.path.join(os.path.split(__file__)[0],'../tiling')
    #t=Table.read(os.path.join(classlocation,'ZTF_quadrantcenters.dat'),
    #             format='ascii.commented_header')
    # Cartesian coordinates in the plane of projection
    #quadrant_x=t['DX']*u.deg
    #quadrant_y=t['DY']*u.deg

    def __init__(self, RA, Dec, number=None, missing=None):
        """
        z=ZTFtile(RA, Dec, number, missing)
        creates a ZTFtile object centered at the given coordinates
        coordinates can be Quantities (with units) or floats (assumed degrees)
        number is just used for informational purposes (bookkeeping)
        missing lists the missing quadrants (potentially None or [])
        if missing is not None, should be a list of quadrants (0--63) or CCD,quadrant pairs (1-16,1-4)
        they will be removed from the quadrant-by-quadrant WCS

        """
        quad = self.get_quadrant_centers()
        self.quadrant_x=quad[:,2]*u.deg
        self.quadrant_y=quad[:,3]*u.deg
        self.quadrant_size=np.array([3072,3080])
        self.quadrant_scale=1*u.arcsec

        if isinstance(RA, astropy.units.quantity.Quantity):            
            self.RA=RA
        else:
            self.RA=RA*u.deg
        if isinstance(Dec, astropy.units.quantity.Quantity):            
            self.Dec=Dec
        else:
            self.Dec=Dec*u.deg

        self.number=number

        self.quadrant_RA,self.quadrant_Dec=self.quadrant_centers()
        
        self.missing_quadrants=[]
        if missing is not None and len(missing)>0:
            for m in missing:
                if isinstance(m,int) or len(m)==1:
                    # assume this is 0 origin
                    self.missing_quadrants.append(m)
                elif len(m)==2:
                    # this assumes CCD and quadrant numbers are 1-origin
                    self.missing_quadrants.append(4*(m[0]-1)+(m[1]-1))
        good_quadrants=np.setdiff1d(np.arange(len(self.quadrant_RA)),
                                    np.array(self.missing_quadrants))
        
        self._wcs=[]
        for i in good_quadrants:
            self._wcs.append(self.quadrant_WCS(self.quadrant_RA[i],self.quadrant_Dec[i]))

    def quadrant_WCS(self, quadrant_RA, quadrant_Dec):
        """
        w=quadrant_WCS(quadrant_RA, quadrant_Dec)
        returns a WCS object that is specific to the quadrant RA and Dec specified
        overall scale, size are determined by class variables
        """
        
        w=WCS(naxis=2)
        w.wcs.crpix=(self.quadrant_size+1.0)/2
        w.wcs.ctype=['RA---TAN','DEC--TAN']
        w.wcs.crval=[quadrant_RA.value,
                     max(min(quadrant_Dec.value,90),-90)]
        w.wcs.cd=[[-self.quadrant_scale.to(u.deg).value,0],
                  [0,-self.quadrant_scale.to(u.deg).value]]
        return w

    def quadrant_centers(self):
        """
        alpha,delta=quadrant_centers()
        return celestial coordinates for the centers of each quadrant
        given the pointing center RA,Dec
        """
        # convert to Native longitude (phi) and latitude (theta)
        # need intermediate step (Rtheta) to deal with TAN projection
        # Calabretta & Greisen (2002), Eqn. 14,15
        phi=np.arctan2(self.quadrant_x,-self.quadrant_y)
        Rtheta=np.sqrt(self.quadrant_x**2+self.quadrant_y**2)
        # now to theta using the TAN projection
        # Calabrett & Greisen (2002), Eqn. 55
        theta=np.arctan2(1,Rtheta.to(u.rad).value)*u.rad
        
        # central point of projection (native pole)
        alphap=self.RA
        deltap=self.Dec
        # Native longitude/latitude of fiducial point
        # appropriate for zenithal projections including TAN
        phi0=0*u.deg
        theta0=90*u.deg
        # Native longitude/latitue of celestial pole
        # for delta0<theta0 then phip should be 180 deg
        # and theta0=90 deg
        phip=180*u.deg
        thetap=self.Dec
        
        # Celestial longitude/latitude
        # Calabretta & Greisen (2002), Eqn. 2
        alpha=alphap+np.arctan2(-np.cos(theta)*np.sin(phi-phip),
                                np.sin(theta)*np.cos(deltap)-np.cos(theta)*np.sin(deltap)*np.cos(phi-phip))
        
        delta=(np.arcsin(np.sin(theta)*np.sin(deltap)+np.cos(theta)*np.cos(deltap)*np.cos(phi-phip))).to(u.deg)
        alpha[alpha<0*u.deg]+=360*u.deg
        return alpha,delta

    def get_quadrant_centers(self):
        quad = [[1, 1, 3.2210639743885596, -2.4437146770476383],
                [1, 2, 2.3535001501796575, -2.43881516574295],
                [1, 3, 2.3503417727072713, -3.3081079798782698],
                [1, 4, 3.2184995347493874, -3.3135486187697305],
                [2, 1, 1.360399735814718, -2.4401317085420295],
                [2, 2, 0.4945677283623324, -2.43770260896455],
                [2, 3, 0.49257586768271355, -3.3063833103517815],
                [2, 4, 1.3589823960224123, -3.308995989160921],
                [3, 1, -0.4933901562131375, -2.434813777659866],
                [3, 2, -1.359168612669584, -2.4315618710281117],
                [3, 3, -1.3633930406282657, -3.3004078547524585],
                [3, 4, -0.4970843882478723, -3.3035234264684044],
                [4, 1, -2.3550841263512168, -2.4300020287485142],
                [4, 2, -3.2225696141613875, -2.4288711155723512],
                [4, 3, -3.225987817653737, -3.2986410207627994],
                [4, 4, -2.3578968641038385, -3.29921741067662],
                [5, 1, 3.2270472994872974, -0.5260446790890496],
                [5, 2, 2.360048851811889, -0.5236056946633051],
                [5, 3, 2.3580069857424673, -1.391013438779468],
                [5, 4, 3.2252436129820534, -1.3940385615739272],
                [6, 1, 1.3636942887667292, -0.5244366494318076],
                [6, 2, 0.4985661268688817, -0.5224684314946721],
                [6, 3, 0.4966475798958921, -1.3892521474803983],
                [6, 4, 1.3619833350672432, -1.3914104228161555],
                [7, 1, -0.4885352259899877, -0.5165091052597491],
                [7, 2, -1.353711077416428, -0.5152297582308262],
                [7, 3, -1.3552233617363139, -1.3821691302104016],
                [7, 4, -0.489935604981924, -1.3833014709475784],
                [8, 1, -2.35112039116874, -0.5124206296644447],
                [8, 2, -3.2179755018158858, -0.5120272575149895],
                [8, 3, -3.218910321057886, -1.3798304494175564],
                [8, 4, -2.3518957475858957, -1.3797479763456497],
                [9, 1, 3.2315748392794528, 1.3889779258968369],
                [9, 2, 2.364458464672072, 1.3900361847270215],
                [9, 3, 2.362357247822555, 0.5226598175841949],
                [9, 4, 3.229340886555462, 0.5210066079569343],
                [10, 1, 1.3683735895823672, 1.39427691039317],
                [10, 2, 0.5029965829034689, 1.3958587387612276],
                [10, 3, 0.5011665116661931, 0.5290203277616491],
                [10, 4, 1.366338951342555, 0.5272736507052246],
                [11, 1, -0.4822088351788257, 1.399285270694412],
                [11, 2, -1.347565895275316, 1.4010754421142027],
                [11, 3, -1.3489171520377228, 0.5341040741152838],
                [11, 4, -0.4837320077661326, 0.5324788914730108],
                [12, 1, -2.3424336622973194, 1.4040639060045907],
                [12, 2, -3.2095684182672755, 1.4052262304759773],
                [12, 3, -3.2095018762120544, 0.5372847732203613],
                [12, 4, -2.342590443769107, 0.5366672655753082],
                [13, 1, 3.2399153631397453, 3.309526119435319],
                [13, 2, 2.3717429700729746, 3.3104062851928253],
                [13, 3, 2.3685575187354933, 2.441059229055063],
                [13, 4, 3.2361524599183253, 2.4396874072033157],
                [14, 1, 1.3776907334001645, 3.309296478848194],
                [14, 2, 0.511265387229377, 3.3129964364715527],
                [14, 3, 0.507007945159137, 2.444240399708313],
                [14, 4, 1.372808010956761, 2.440396042373973],
                [15, 1, -0.47857144910010013, 3.31541767888218],
                [15, 2, -1.3449638389147003, 3.3156973843720072],
                [15, 3, -1.3440992664923928, 2.4467894221776447],
                [15, 4, -0.4782503630964142, 2.446667614844775],
                [16, 1, -2.3396869448373576, 3.3204881263460355],
                [16, 2, -3.2078226835720773, 3.3235070629897],
                [16, 3, -3.2079494303333775, 2.453629973388401],
                [16, 4, -2.3403853572098905, 2.4512062297969988]]
        return np.array(quad)

class QuadProb:
    '''
    Class :: Instantiate a QuadProb object that will allow us to calculate the
             probability content in a single quadrant.
    '''
    def __init__(self, RA, Dec):
        self.RA = RA
        self.Dec = Dec
        self.quadMap = {1: [0, 1, 2, 3], 2: [4, 5, 6, 7],
                        3: [8, 9, 10, 11], 4: [12, 13, 14, 15],
                        5: [16, 17, 18, 19], 6: [20, 21, 22, 23],
                        7: [24, 25, 26, 27], 8: [28, 29, 30, 31],
                        9: [32, 33, 34, 35], 10: [36, 37, 38, 39],
                        11: [40, 41, 42, 43], 12: [44, 45, 46, 47],
                        13: [48, 49, 50, 51], 14: [52, 53, 54, 55],
                        15: [56, 57, 58, 59], 16: [60, 61, 62, 63]}
        self.quadrant_size = np.array([3072, 3080])
        self.quadrant_scale = 1*u.arcsec

    def getWCS(self, quad_cents_RA, quad_cents_Dec):
        _wcs = WCS(naxis=2)
        _wcs.wcs.crpix = (self.quadrant_size + 1.0)/2
        _wcs.wcs.ctype = ['RA---TAN', 'DEC--TAN']
        wrapped_quad_cent_Dec = max(min(quad_cents_Dec.value, 90), -90)
        _wcs.wcs.crval = [quad_cents_RA.value, wrapped_quad_cent_Dec]
        _wcs.wcs.cd = [[self.quadrant_scale.to(u.deg).value, 0],
                       [0, -self.quadrant_scale.to(u.deg).value]]
        return _wcs

def get_ztf_quadrants():
    """Calculate ZTF quadrant footprints as offsets from the telescope
    boresight."""
    quad_prob = QuadProb(0, 0)
    ztf_tile = ZTFtile(0, 0)
    quad_cents_ra, quad_cents_dec = ztf_tile.quadrant_centers()
    offsets = np.asarray([
        quad_prob.getWCS(
            quad_cents_ra[quadrant_id],
            quad_cents_dec[quadrant_id]
        ).calc_footprint(axes=quad_prob.quadrant_size)
        for quadrant_id in range(64)])
    return np.transpose(offsets, (2, 0, 1))

def get_quadrant_ipix(nside, ra, dec, subfield_ids=None):

    quadrant_coords = get_ztf_quadrants()

    skyoffset_frames = SkyCoord(ra, dec, unit=u.deg).skyoffset_frame()
    quadrant_coords_icrs = SkyCoord(
                    *np.tile(
                        quadrant_coords[:, np.newaxis, ...],
                        (1, 1, 1)), unit=u.deg,
                    frame=skyoffset_frames[:, np.newaxis, np.newaxis]
                ).transform_to(ICRS)
    quadrant_xyz = np.moveaxis(
        quadrant_coords_icrs.cartesian.xyz.value, 0, -1)[0]

    ipixs = []
    for subfield_id, xyz in enumerate(quadrant_xyz):
        if not subfield_ids is None:
            if not subfield_id in subfield_ids:
                continue
        ipix = hp.query_polygon(nside, xyz)
        ipixs.append(ipix.tolist())
    return ipixs




"""
Database schema.
"""

from flask_sqlalchemy import SQLAlchemy
import pkg_resources
from gwemopt.flaskapp import app
import numpy as np
import healpy as hp
from astropy.table import Table
from astropy.io import ascii

db = SQLAlchemy(app)

def create_all(catalogFile):
    db.drop_all()
    db.create_all()

    nside = 256

    cat = ascii.read(catalogFile,format='ecsv')
    ipixs = hp.ang2pix(nside, cat['RAJ2000'].to('deg').value, cat['DEJ2000'].to('deg').value, lonlat=True)

    for ii,row in enumerate(cat):
        if np.mod(ii,10000) == 0:
            print('%d/%d'%(ii,len(cat)))
        ra, dec, dist = float(row["RAJ2000"]), float(row["DEJ2000"]), float(row["Dist"])
        Bmag, Kmag = float(row["Bmag"]), float(row["Kmag2"])
        ipix = ipixs[ii]

        db.session.merge(Galaxy(galaxy_id=int(ii),
                               ra=ra, dec=dec, dist=dist,
                               Bmag=Bmag, Kmag=Kmag,
                               ipix=int(ipix)))
    db.session.commit()

def read_catalog():

    galaxies = Galaxy.query.all()
    cat = Table(names=('RAJ2000', 'DEJ2000', 'Dist', 'Bmag', 'Kmag', 'ipix'), dtype=('f8', 'f8', 'f8', 'f8', 'f8', 'i8'))
    for galaxy in galaxies:
        gal = [galaxy.ra, galaxy.dec, galaxy.dist, galaxy.Bmag, galaxy.Kmag, galaxy.ipix]
        cat.add_row(gal)

class Galaxy(db.Model):
    """Galaxy information"""

    galaxy_id = db.Column(
        db.Integer,
        primary_key=True,
        comment='Galaxy field ID')

    ra = db.Column(
        db.Float,
        nullable=False,
        comment='RA of galaxy')

    dec = db.Column(
        db.Float,
        nullable=False,
        comment='Dec of galaxy')

    dist = db.Column(
        db.Float,
        nullable=False,
        comment='Distance of galaxy')

    Bmag = db.Column(
        db.Float,
        nullable=False,
        comment='B band mag of galaxy')

    Kmag = db.Column(
        db.Float,
        nullable=False,
        comment='K band mag of galaxy')

    ipix = db.Column(
        db.Integer,
        comment='Healpix index')



import datetime
import os

from flask import Flask
from flask_humanize import Humanize
from werkzeug.routing import BaseConverter

# Application object
app = Flask(__name__, instance_relative_config=True)
app.config['DEBUG'] = True
app.config['SQLALCHEMY_DATABASE_URI'] = 'postgresql:///gwemopt'

app.config['TEMPLATES_AUTO_RELOAD'] = True

# Turn off memory-intensive modification tracking.
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

# Default secret key: secure and random. However, sessions are not preserved
# across different Python processes.
app.config['SECRET_KEY'] = os.urandom(24)

# Celery configuration.
# Use pickle serializer, because it supports byte values.
# Use redis broker, because it supports locks (and thus singleton tasks).
app.config['CELERY_BROKER_URL'] = 'redis://'
app.config['CELERY_ACCEPT_CONTENT'] = ['json', 'pickle']
app.config['CELERY_EVENT_SERIALIZER'] = 'json'
app.config['CELERY_RESULT_SERIALIZER'] = 'pickle'
app.config['CELERY_TASK_SERIALIZER'] = 'pickle'

# Apply instance config.
#app.config.from_pyfile('application.cfg', silent=False)


class DateTimeConverter(BaseConverter):

    def to_python(self, value):
        try:
            return datetime.datetime.strptime(value, '%y%m%d').date()
        except ValueError:
            return datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')

    def to_url(self, value):
        return value.isoformat(timespec='seconds')


app.url_map.converters['datetime'] = DateTimeConverter

app.jinja_env.globals['now'] = datetime.datetime.now

humanize = Humanize(app)



import os
from astropy import time
import astropy.units as u
from astropy.table import Table
from celery.task import PeriodicTask
from celery.utils.log import get_task_logger
from celery.local import PromiseProxy
import numpy as np
import pyvo.dal
import requests

client = pyvo.dal.TAPService('https://irsa.ipac.caltech.edu/TAP',)

def ztf_references():
    refstable = client.search("""
    SELECT field, fid, maglimit FROM ztf.ztf_current_meta_ref
    WHERE (nframes >= 15) AND (startobsdate >= '2018-02-05T00:00:00Z')
    AND (field < 2000)
    """).to_table()

    refs = refstable.group_by(['field', 'fid']).groups.aggregate(np.mean)
    refs = refs.filled()

    refs_grouped_by_field = refs.group_by('field').groups
    for field_id, rows in zip(refs_grouped_by_field.keys,
                              refs_grouped_by_field):
        print(field_id, rows)

ztf_references()
 




import os, sys
import copy
import numpy as np
import healpy as hp
import itertools

from scipy.stats import norm

import astropy.coordinates
from astropy.time import Time, TimeDelta
import astropy.units as u

def read_transients(params, map_struct):

    nside = params["nside"]
    prob_data_sorted = np.sort(map_struct["prob"])[::-1]
    prob_data_indexes = np.argsort(map_struct["prob"])[::-1]
    prob_data_cumsum = np.cumsum(prob_data_sorted)

    lines = [line.rstrip('\n') for line in open(params["transientsFile"])]
    lines = lines[1:]
    lines = filter(None,lines)

    transients_struct = {}
    transients_struct["data"] = np.empty((0,8))
    transients_struct["name"] = []
    transients_struct["classification"] = []

    for line in lines:
        line = line.replace('"','')
        lineSplit = line.split(",")
        if lineSplit[0] == "NULL":
            continue

        name = lineSplit[0]
        ra = float(lineSplit[2])
        dec = float(lineSplit[3])
        classification = lineSplit[6]

        if not lineSplit[11] == "":
            mag = float(lineSplit[11])
        else:
            mag = -1

        if not lineSplit[13] == "":
            z = float(lineSplit[13])
        else:
            z = -1
        if not lineSplit[14] == "":
            mpc = float(lineSplit[14])
        else:
            mpc = -1

        if not lineSplit[15] == "":
            discovery_mjd = Time(lineSplit[15], format='isot').mjd
        else:
            discovery_mjd = -1

        if not lineSplit[16] == "":
            nondetection_mjd = Time(lineSplit[16], format='isot').mjd
        else:
            #nondetection_mjd = -1
            nondetection_mjd = discovery_mjd

        event_mjd = Time(params["gpstime"], format='gps', scale='utc').mjd
        if (event_mjd < nondetection_mjd) or (event_mjd > discovery_mjd + params["dt"]):
            continue

        ipix = hp.ang2pix(nside, theta=ra, phi=dec, lonlat=True)
        cumprob = prob_data_cumsum[ipix]

        transients_struct["data"] = np.append(transients_struct["data"],np.array([[ra,dec,mag,discovery_mjd,nondetection_mjd,z,mpc,cumprob]]),axis=0)
        transients_struct["name"].append(name)
        transients_struct["classification"].append(classification)

    transients_struct["name"] = np.array(transients_struct["name"])
    transients_struct["classification"] = np.array(transients_struct["classification"])

    idx = np.argsort(transients_struct["data"][:,7])
    transients_struct["data"] = transients_struct["data"][idx,:]
    transients_struct["name"] = transients_struct["name"][idx]
    transients_struct["classification"] = transients_struct["classification"][idx]

    return transients_struct


